'''

$Author: murco $
$Id: Processor.py 1 2012-11-12 18:56:07Z murco $
'''

from .jroproc_voltage import *
from .jroproc_spectra import *
from .jroproc_heispectra import *
from .jroproc_amisr import *
from .jroproc_correlation import *
from .jroproc_parameters import *
from .jroproc_spectra_lags import *
from .jroproc_spectra_acf import *
from .bltrproc_parameters import *
from .pxproc_parameters import *



'''
@author: roj-idl71
'''
#USED IN jroplot_spectra.py
RTI_CODE = 0            #Range time intensity (RTI).
SPEC_CODE = 1           #Spectra (and Cross-spectra) information.
CROSS_CODE = 2          #Cross-Correlation information.
COH_CODE = 3            #Coherence map.
BASE_CODE = 4           #Base lines graphic.
ROW_CODE = 5            #Row Spectra.
TOTAL_CODE = 6          #Total Power.
DRIFT_CODE = 7          #Drifts graphics.
HEIGHT_CODE = 8         #Height profile.
PHASE_CODE = 9          #Signal Phase.

POWER_CODE = 16
NOISE_CODE = 17
BEACON_CODE = 18

#USED IN jroplot_parameters.py
WIND_CODE = 22
MSKYMAP_CODE = 23
MPHASE_CODE = 24

MOMENTS_CODE = 25
PARMS_CODE = 26
SPECFIT_CODE = 27
EWDRIFT_CODE = 28



'''
Created on Set 9, 2015

@author: roj-idl71 Karim Kuyeng
'''

import os
import sys
import glob
import fnmatch
import datetime
import time
import re
import h5py
import numpy

try:
    from gevent import sleep
except:
    from time import sleep

from schainpy.model.data.jroheaderIO import RadarControllerHeader, SystemHeader
from schainpy.model.data.jrodata import Voltage
from schainpy.model.proc.jroproc_base import ProcessingUnit, Operation
from numpy import imag

class AMISRReader(ProcessingUnit):
    '''
    classdocs
    '''

    def __init__(self):
        '''
        Constructor
        '''
        
        ProcessingUnit.__init__(self)
        
        self.set = None
        self.subset = None
        self.extension_file = '.h5'
        self.dtc_str = 'dtc'
        self.dtc_id = 0
        self.status = True
        self.isConfig = False
        self.dirnameList = []
        self.filenameList = []
        self.fileIndex = None
        self.flagNoMoreFiles = False
        self.flagIsNewFile = 0
        self.filename = ''
        self.amisrFilePointer = None
        
        
        self.dataset = None
        
        
       

        self.profileIndex = 0
        
        
        self.beamCodeByFrame = None
        self.radacTimeByFrame = None
        
        self.dataset = None
        
             
                
        
        self.__firstFile = True
        
        self.buffer = None
        
        
        self.timezone = 'ut'
        
        self.__waitForNewFile = 20
        self.__filename_online = None 
        #Is really necessary create the output object in the initializer
        self.dataOut = Voltage()
        
    def setup(self,path=None,
                    startDate=None, 
                    endDate=None, 
                    startTime=None, 
                    endTime=None,
                    walk=True,
                    timezone='ut',
                    all=0,
                    code = None,
                    nCode = 0,
                    nBaud = 0,
                    online=False):
    
        self.timezone = timezone
        self.all = all
        self.online = online
        
        self.code = code
        self.nCode = int(nCode)
        self.nBaud = int(nBaud)
        
        
        
        #self.findFiles()
        if not(online):
            #Busqueda de archivos offline
            self.searchFilesOffLine(path, startDate, endDate, startTime, endTime, walk)
        else:
            self.searchFilesOnLine(path, startDate, endDate, startTime,endTime,walk)
        
        if not(self.filenameList):
            print("There is no files into the folder: %s"%(path))
                
            sys.exit(-1)
            
        self.fileIndex = -1
        
        self.readNextFile(online)   
               
        '''
        Add code
        '''        
        self.isConfig = True
        
        pass
    
    
    def readAMISRHeader(self,fp):
        header = 'Raw11/Data/RadacHeader'
        self.beamCodeByPulse = fp.get(header+'/BeamCode') # LIST OF BEAMS PER PROFILE, TO BE USED ON REARRANGE
        self.beamCode = fp.get('Raw11/Data/Beamcodes') # NUMBER OF CHANNELS AND IDENTIFY POSITION TO CREATE A FILE WITH THAT INFO
        #self.code = fp.get(header+'/Code') # NOT USE FOR THIS
        self.frameCount = fp.get(header+'/FrameCount')# NOT USE FOR THIS
        self.modeGroup = fp.get(header+'/ModeGroup')# NOT USE FOR THIS
        self.nsamplesPulse = fp.get(header+'/NSamplesPulse')# TO GET NSA OR USING DATA FOR THAT
        self.pulseCount = fp.get(header+'/PulseCount')# NOT USE FOR THIS
        self.radacTime = fp.get(header+'/RadacTime')# 1st TIME ON FILE ANDE CALCULATE THE REST WITH IPP*nindexprofile
        self.timeCount = fp.get(header+'/TimeCount')# NOT USE FOR THIS
        self.timeStatus = fp.get(header+'/TimeStatus')# NOT USE FOR THIS
        self.rangeFromFile = fp.get('Raw11/Data/Samples/Range')
        self.frequency =  fp.get('Rx/Frequency')
        txAus = fp.get('Raw11/Data/Pulsewidth')
         
        
        self.nblocks = self.pulseCount.shape[0] #nblocks
       
        self.nprofiles = self.pulseCount.shape[1] #nprofile
        self.nsa = self.nsamplesPulse[0,0] #ngates
        self.nchannels = self.beamCode.shape[1]
        self.ippSeconds = (self.radacTime[0][1] -self.radacTime[0][0]) #Ipp in seconds
        #self.__waitForNewFile = self.nblocks  # wait depending on the number of blocks since each block is 1 sec
        self.__waitForNewFile = self.nblocks * self.nprofiles * self.ippSeconds # wait until new file is created
        
        #filling radar controller header parameters
        self.__ippKm = self.ippSeconds *.15*1e6 # in km
        self.__txA = (txAus.value)*.15 #(ipp[us]*.15km/1us) in km
        self.__txB = 0
        nWindows=1
        self.__nSamples = self.nsa 
        self.__firstHeight = self.rangeFromFile[0][0]/1000 #in km
        self.__deltaHeight = (self.rangeFromFile[0][1] - self.rangeFromFile[0][0])/1000 
        
        #for now until understand why the code saved is different (code included even though code not in tuf file)
        #self.__codeType = 0
       # self.__nCode = None
       # self.__nBaud = None
        self.__code = self.code
        self.__codeType = 0
        if self.code != None:
            self.__codeType = 1
        self.__nCode = self.nCode
        self.__nBaud = self.nBaud
        #self.__code = 0
        
        #filling system header parameters
        self.__nSamples = self.nsa
        self.newProfiles = self.nprofiles/self.nchannels 
        self.__channelList = list(range(self.nchannels))
        
        self.__frequency = self.frequency[0][0]
        

    
    def createBuffers(self):
    
        pass      
       
    def __setParameters(self,path='', startDate='',endDate='',startTime='', endTime='', walk=''):
        self.path = path
        self.startDate = startDate
        self.endDate = endDate
        self.startTime = startTime
        self.endTime = endTime
        self.walk = walk
    
    def __checkPath(self):
        if os.path.exists(self.path):
            self.status = 1
        else:
            self.status = 0
            print('Path:%s does not exists'%self.path)
            
        return
    
    
    def __selDates(self, amisr_dirname_format):
        try:
            year = int(amisr_dirname_format[0:4])
            month = int(amisr_dirname_format[4:6])
            dom = int(amisr_dirname_format[6:8])
            thisDate = datetime.date(year,month,dom)
            
            if (thisDate>=self.startDate and thisDate <= self.endDate):
                return amisr_dirname_format
        except:
            return None
    
    
    def __findDataForDates(self,online=False):
        
        if not(self.status):
            return None
        
        pat = '\d+.\d+'
        dirnameList = [re.search(pat,x) for x in os.listdir(self.path)]
        dirnameList = [x for x in dirnameList if x!=None]
        dirnameList = [x.string for x in dirnameList]
        if not(online):
            dirnameList = [self.__selDates(x) for x in dirnameList]
            dirnameList = [x for x in dirnameList if x!=None]
        if len(dirnameList)>0:
            self.status = 1
            self.dirnameList = dirnameList
            self.dirnameList.sort()
        else:
            self.status = 0
            return None
    
    def __getTimeFromData(self):
        startDateTime_Reader = datetime.datetime.combine(self.startDate,self.startTime)
        endDateTime_Reader = datetime.datetime.combine(self.endDate,self.endTime)

        print('Filtering Files from %s to %s'%(startDateTime_Reader, endDateTime_Reader))
        print('........................................')
        filter_filenameList = []
        self.filenameList.sort()
        #for i in range(len(self.filenameList)-1):
        for i in range(len(self.filenameList)):
            filename = self.filenameList[i]
            fp = h5py.File(filename,'r')
            time_str = fp.get('Time/RadacTimeString')
            
            startDateTimeStr_File = time_str[0][0].split('.')[0]
            junk = time.strptime(startDateTimeStr_File, '%Y-%m-%d %H:%M:%S')
            startDateTime_File = datetime.datetime(junk.tm_year,junk.tm_mon,junk.tm_mday,junk.tm_hour, junk.tm_min, junk.tm_sec)
            
            endDateTimeStr_File = time_str[-1][-1].split('.')[0]
            junk = time.strptime(endDateTimeStr_File, '%Y-%m-%d %H:%M:%S')
            endDateTime_File = datetime.datetime(junk.tm_year,junk.tm_mon,junk.tm_mday,junk.tm_hour, junk.tm_min, junk.tm_sec)
            
            fp.close()
            
            if self.timezone == 'lt':
                startDateTime_File = startDateTime_File - datetime.timedelta(minutes = 300)
                endDateTime_File = endDateTime_File - datetime.timedelta(minutes = 300)

            if (endDateTime_File>=startDateTime_Reader and endDateTime_File<endDateTime_Reader):
                #self.filenameList.remove(filename)
                filter_filenameList.append(filename)
            
            if (endDateTime_File>=endDateTime_Reader):
                break
            
        
        filter_filenameList.sort()
        self.filenameList = filter_filenameList
        return 1
    
    def __filterByGlob1(self, dirName):
        filter_files = glob.glob1(dirName, '*.*%s'%self.extension_file)
        filter_files.sort()
        filterDict = {}
        filterDict.setdefault(dirName)
        filterDict[dirName] = filter_files
        return filterDict
    
    def __getFilenameList(self, fileListInKeys, dirList):
        for value in fileListInKeys:
            dirName = list(value.keys())[0]
            for file in value[dirName]:
                filename = os.path.join(dirName, file)
                self.filenameList.append(filename)
    
    
    def __selectDataForTimes(self, online=False):
        #aun no esta implementado el filtro for tiempo
        if not(self.status):
            return None
        
        dirList = [os.path.join(self.path,x) for x in self.dirnameList]
        
        fileListInKeys = [self.__filterByGlob1(x) for x in dirList]
        
        self.__getFilenameList(fileListInKeys, dirList)
        if not(online):
            #filtro por tiempo
            if not(self.all):
                self.__getTimeFromData()

            if len(self.filenameList)>0:
                self.status = 1
                self.filenameList.sort()
            else:
                self.status = 0
                return None
            
        else:
            #get the last file - 1
            self.filenameList = [self.filenameList[-2]]
        
        new_dirnameList = []
        for dirname in self.dirnameList:
            junk = numpy.array([dirname in x for x in self.filenameList])
            junk_sum = junk.sum()
            if junk_sum > 0:
                new_dirnameList.append(dirname)
        self.dirnameList = new_dirnameList
        return 1
    
    def searchFilesOnLine(self, path, startDate, endDate, startTime=datetime.time(0,0,0),
                            endTime=datetime.time(23,59,59),walk=True):
        
        if endDate ==None:
         startDate = datetime.datetime.utcnow().date()
         endDate = datetime.datetime.utcnow().date()
        
        self.__setParameters(path=path, startDate=startDate, endDate=endDate,startTime = startTime,endTime=endTime, walk=walk)
        
        self.__checkPath()
        
        self.__findDataForDates(online=True)
        
        self.dirnameList = [self.dirnameList[-1]]
        
        self.__selectDataForTimes(online=True)
        
        return
        
    
    def searchFilesOffLine(self,
                            path,
                            startDate,
                            endDate,
                            startTime=datetime.time(0,0,0),
                            endTime=datetime.time(23,59,59),
                            walk=True):
        
        self.__setParameters(path, startDate, endDate, startTime, endTime, walk)
        
        self.__checkPath()
        
        self.__findDataForDates()
        
        self.__selectDataForTimes()
        
        for i in range(len(self.filenameList)):
            print("%s" %(self.filenameList[i]))
        
        return 
        
    def __setNextFileOffline(self):
        idFile = self.fileIndex

        while (True):
            idFile += 1
            if not(idFile < len(self.filenameList)):
                self.flagNoMoreFiles = 1
                print("No more Files")
                return 0

            filename = self.filenameList[idFile]

            amisrFilePointer = h5py.File(filename,'r')
            
            break

        self.flagIsNewFile = 1
        self.fileIndex = idFile
        self.filename = filename

        self.amisrFilePointer = amisrFilePointer

        print("Setting the file: %s"%self.filename)

        return 1
    
    
    def __setNextFileOnline(self):
        filename = self.filenameList[0]
        if self.__filename_online != None:
            self.__selectDataForTimes(online=True)
            filename = self.filenameList[0]
            wait = 0
            while self.__filename_online == filename:
                print('waiting %d seconds to get a new file...'%(self.__waitForNewFile))
                if wait == 5:
                    return 0
                sleep(self.__waitForNewFile)
                self.__selectDataForTimes(online=True)
                filename = self.filenameList[0]
                wait += 1
        
        self.__filename_online = filename
        
        self.amisrFilePointer = h5py.File(filename,'r')
        self.flagIsNewFile = 1
        self.filename = filename
        print("Setting the file: %s"%self.filename)
        return 1
    
    
    def readData(self):
        buffer = self.amisrFilePointer.get('Raw11/Data/Samples/Data')
        re = buffer[:,:,:,0]
        im = buffer[:,:,:,1]
        dataset = re + im*1j
        self.radacTime = self.amisrFilePointer.get('Raw11/Data/RadacHeader/RadacTime')
        timeset = self.radacTime[:,0]
        return dataset,timeset
    
    def reshapeData(self):
    #self.beamCodeByPulse, self.beamCode, self.nblocks, self.nprofiles, self.nsa, 
        channels = self.beamCodeByPulse[0,:]
        nchan = self.nchannels
        #self.newProfiles = self.nprofiles/nchan #must be defined on filljroheader
        nblocks = self.nblocks
        nsamples = self.nsa
    
        #Dimensions : nChannels, nProfiles, nSamples
        new_block = numpy.empty((nblocks, nchan, self.newProfiles, nsamples), dtype="complex64")
        ############################################
    
        for thisChannel in range(nchan):
            new_block[:,thisChannel,:,:] = self.dataset[:,numpy.where(channels==self.beamCode[0][thisChannel])[0],:]

        
        new_block = numpy.transpose(new_block, (1,0,2,3))
        new_block = numpy.reshape(new_block, (nchan,-1, nsamples))
        
        return new_block 
    
    def updateIndexes(self):
    
        pass
        
    def fillJROHeader(self):
        
        #fill radar controller header
        self.dataOut.radarControllerHeaderObj = RadarControllerHeader(ippKm=self.__ippKm,
                                                                      txA=self.__txA,
                                                                      txB=0,
                                                                      nWindows=1,
                                                                      nHeights=self.__nSamples,
                                                                      firstHeight=self.__firstHeight,
                                                                      deltaHeight=self.__deltaHeight,
                                                                      codeType=self.__codeType,
                                                                      nCode=self.__nCode, nBaud=self.__nBaud,
                                                                      code = self.__code,
                                                                      fClock=1)
    
        
        
        #fill system header
        self.dataOut.systemHeaderObj = SystemHeader(nSamples=self.__nSamples,
                                                    nProfiles=self.newProfiles,
                                                    nChannels=len(self.__channelList),
                                                    adcResolution=14,
                                                    pciDioBusWith=32)
        
        self.dataOut.type = "Voltage"
        
        self.dataOut.data = None
        
        self.dataOut.dtype = numpy.dtype([('real','<i8'),('imag','<i8')])
        
#        self.dataOut.nChannels = 0
        
#        self.dataOut.nHeights = 0
        
        self.dataOut.nProfiles = self.newProfiles*self.nblocks
        
        #self.dataOut.heightList = self.__firstHeigth + numpy.arange(self.__nSamples, dtype = numpy.float)*self.__deltaHeigth
        ranges = numpy.reshape(self.rangeFromFile.value,(-1))
        self.dataOut.heightList =  ranges/1000.0 #km
        
        
        self.dataOut.channelList = self.__channelList
        
        self.dataOut.blocksize = self.dataOut.getNChannels() * self.dataOut.getNHeights()
        
#        self.dataOut.channelIndexList = None
        
        self.dataOut.flagNoData = True
        
        #Set to TRUE if the data is discontinuous 
        self.dataOut.flagDiscontinuousBlock = False
        
        self.dataOut.utctime = None
         
        #self.dataOut.timeZone = -5 #self.__timezone/60  #timezone like jroheader, difference in minutes between UTC and localtime
        if self.timezone == 'lt':
            self.dataOut.timeZone = time.timezone / 60. #get the timezone in minutes
        else: 
            self.dataOut.timeZone = 0 #by default time is UTC

        self.dataOut.dstFlag = 0
        
        self.dataOut.errorCount = 0
        
        self.dataOut.nCohInt = 1
        
        self.dataOut.flagDecodeData = False #asumo que la data esta decodificada
    
        self.dataOut.flagDeflipData = False #asumo que la data esta sin flip
        
        self.dataOut.flagShiftFFT = False
        
        self.dataOut.ippSeconds = self.ippSeconds
        
        #Time interval between profiles 
        #self.dataOut.timeInterval = self.dataOut.ippSeconds * self.dataOut.nCohInt
        
        self.dataOut.frequency = self.__frequency
        
        self.dataOut.realtime = self.online
        pass
    
    def readNextFile(self,online=False):
        
        if not(online):
            newFile = self.__setNextFileOffline()
        else:
            newFile = self.__setNextFileOnline() 
        
        if not(newFile):
            return 0
        
        #if self.__firstFile:
        self.readAMISRHeader(self.amisrFilePointer)
        self.createBuffers()
        self.fillJROHeader()
        #self.__firstFile = False
            
        
        
        self.dataset,self.timeset = self.readData()
        
        if self.endDate!=None:
         endDateTime_Reader = datetime.datetime.combine(self.endDate,self.endTime)
         time_str = self.amisrFilePointer.get('Time/RadacTimeString')
         startDateTimeStr_File = time_str[0][0].split('.')[0]
         junk = time.strptime(startDateTimeStr_File, '%Y-%m-%d %H:%M:%S')
         startDateTime_File = datetime.datetime(junk.tm_year,junk.tm_mon,junk.tm_mday,junk.tm_hour, junk.tm_min, junk.tm_sec)
         if self.timezone == 'lt':
          startDateTime_File = startDateTime_File - datetime.timedelta(minutes = 300)
         if (startDateTime_File>endDateTime_Reader):
             return 0
        
        self.jrodataset = self.reshapeData()
        #----self.updateIndexes()
        self.profileIndex = 0
        
        return 1
    
    
    def __hasNotDataInBuffer(self):
        if self.profileIndex >= (self.newProfiles*self.nblocks):
            return 1
        return 0
            
            
    def getData(self):
        
        if self.flagNoMoreFiles:
            self.dataOut.flagNoData = True
            return 0
        
        if self.__hasNotDataInBuffer():
            if not (self.readNextFile(self.online)):
                return 0

        
        if self.dataset is None: # setear esta condicion cuando no hayan datos por leers
            self.dataOut.flagNoData = True 
            return 0
        
        #self.dataOut.data = numpy.reshape(self.jrodataset[self.profileIndex,:],(1,-1))
        
        self.dataOut.data = self.jrodataset[:,self.profileIndex,:]
        
        #self.dataOut.utctime = self.jrotimeset[self.profileIndex]
        #verificar basic header de jro data y ver si es compatible con este valor
        #self.dataOut.utctime = self.timeset + (self.profileIndex * self.ippSeconds * self.nchannels)
        indexprof = numpy.mod(self.profileIndex, self.newProfiles)
        indexblock = self.profileIndex/self.newProfiles
        #print indexblock, indexprof
        self.dataOut.utctime = self.timeset[indexblock] + (indexprof * self.ippSeconds * self.nchannels)
        self.dataOut.profileIndex = self.profileIndex
        self.dataOut.flagNoData = False
#         if indexprof == 0:
#             print self.dataOut.utctime
        
        self.profileIndex += 1
        
        return self.dataOut.data
    
       
    def run(self, **kwargs):
        '''
        This method will be called many times so here you should put all your code
        '''
        
        if not self.isConfig:
            self.setup(**kwargs)
            self.isConfig = True
            
        self.getData()



import os
import sys
import glob
import fnmatch
import datetime
import time
import re
import h5py
import numpy

from scipy.optimize import curve_fit
from scipy import asarray as ar, exp
from scipy import stats

from numpy.ma.core import getdata

SPEED_OF_LIGHT = 299792458
SPEED_OF_LIGHT = 3e8

try:
    from gevent import sleep
except:
    from time import sleep

from schainpy.model.data.jrodata import Spectra
#from schainpy.model.data.BLTRheaderIO import FileHeader, RecordHeader
from schainpy.model.proc.jroproc_base import ProcessingUnit, Operation
#from schainpy.model.io.jroIO_bltr import BLTRReader
from numpy import imag, shape, NaN, empty


class Header(object):

    def __init__(self):
        raise NotImplementedError

    def read(self):

        raise NotImplementedError

    def write(self):

        raise NotImplementedError

    def printInfo(self):

        message = "#" * 50 + "\n"
        message += self.__class__.__name__.upper() + "\n"
        message += "#" * 50 + "\n"

        keyList = list(self.__dict__.keys())
        keyList.sort()

        for key in keyList:
            message += "%s = %s" % (key, self.__dict__[key]) + "\n"

        if "size" not in keyList:
            attr = getattr(self, "size")

            if attr:
                message += "%s = %s" % ("size", attr) + "\n"

        # print message


FILE_HEADER = numpy.dtype([  # HEADER 1024bytes
                          ('Hname', 'a32'),  # Original file name
                          # Date and time when the file was created
                          ('Htime', numpy.str_, 32),
                          # Name of operator who created the file
                          ('Hoper', numpy.str_, 64),
                          # Place where the measurements was carried out
                          ('Hplace', numpy.str_, 128),
                          # Description of measurements
                          ('Hdescr', numpy.str_, 256),
                          ('Hdummy', numpy.str_, 512),  # Reserved space
                          # Main chunk 8bytes
                          # Main chunk signature FZKF or NUIG
                          ('Msign', numpy.str_, 4),
                          ('MsizeData', '<i4'),  # Size of data block main chunk
                          # Processing DSP parameters 36bytes
                          ('PPARsign', numpy.str_, 4),  # PPAR signature
                          ('PPARsize', '<i4'),  # PPAR size of block
                          ('PPARprf', '<i4'),  # Pulse repetition frequency
                          ('PPARpdr', '<i4'),  # Pulse duration
                          ('PPARsft', '<i4'),  # FFT length
                          # Number of spectral (in-coherent) averages
                          ('PPARavc', '<i4'),
                          # Number of lowest range gate for moment estimation
                          ('PPARihp', '<i4'),
                          # Count for gates for moment estimation
                          ('PPARchg', '<i4'),
                          # switch on/off polarimetric measurements. Should be 1.
                          ('PPARpol', '<i4'),
                          # Service DSP parameters 112bytes
                          # STC attenuation on the lowest ranges on/off
                          ('SPARatt', '<i4'),
                          ('SPARtx', '<i4'),  # OBSOLETE
                          ('SPARaddGain0', '<f4'),  # OBSOLETE
                          ('SPARaddGain1', '<f4'),  # OBSOLETE
                          # Debug only. It normal mode it is 0.
                          ('SPARwnd', '<i4'),
                          # Delay between sync pulse and tx pulse for phase corr, ns
                          ('SPARpos', '<i4'),
                          # "add to pulse" to compensate for delay between the leading edge of driver pulse and envelope of the RF signal.
                          ('SPARadd', '<i4'),
                          # Time for measuring txn pulse phase. OBSOLETE
                          ('SPARlen', '<i4'),
                          ('SPARcal', '<i4'),  # OBSOLETE
                          ('SPARnos', '<i4'),  # OBSOLETE
                          ('SPARof0', '<i4'),  # detection threshold
                          ('SPARof1', '<i4'),  # OBSOLETE
                          ('SPARswt', '<i4'),  # 2nd moment estimation threshold
                          ('SPARsum', '<i4'),  # OBSOLETE
                          ('SPARosc', '<i4'),  # flag Oscillosgram mode
                          ('SPARtst', '<i4'),  # OBSOLETE
                          ('SPARcor', '<i4'),  # OBSOLETE
                          ('SPARofs', '<i4'),  # OBSOLETE
                          # Hildebrand div noise detection on noise gate
                          ('SPARhsn', '<i4'),
                          # Hildebrand div noise detection on all gates
                          ('SPARhsa', '<f4'),
                          ('SPARcalibPow_M', '<f4'),  # OBSOLETE
                          ('SPARcalibSNR_M', '<f4'),  # OBSOLETE
                          ('SPARcalibPow_S', '<f4'),  # OBSOLETE
                          ('SPARcalibSNR_S', '<f4'),  # OBSOLETE
                          # Lowest range gate for spectra saving Raw_Gate1 >=5
                          ('SPARrawGate1', '<i4'),
                          # Number of range gates with atmospheric signal
                          ('SPARrawGate2', '<i4'),
                          # flag - IQ or spectra saving on/off
                          ('SPARraw', '<i4'),
                          ('SPARprc', '<i4'), ])  # flag - Moment estimation switched on/off


class FileHeaderMIRA35c(Header):

    def __init__(self):

        self.Hname = None
        self.Htime = None
        self.Hoper = None
        self.Hplace = None
        self.Hdescr = None
        self.Hdummy = None

        self.Msign = None
        self.MsizeData = None

        self.PPARsign = None
        self.PPARsize = None
        self.PPARprf = None
        self.PPARpdr = None
        self.PPARsft = None
        self.PPARavc = None
        self.PPARihp = None
        self.PPARchg = None
        self.PPARpol = None
        # Service DSP parameters
        self.SPARatt = None
        self.SPARtx = None
        self.SPARaddGain0 = None
        self.SPARaddGain1 = None
        self.SPARwnd = None
        self.SPARpos = None
        self.SPARadd = None
        self.SPARlen = None
        self.SPARcal = None
        self.SPARnos = None
        self.SPARof0 = None
        self.SPARof1 = None
        self.SPARswt = None
        self.SPARsum = None
        self.SPARosc = None
        self.SPARtst = None
        self.SPARcor = None
        self.SPARofs = None
        self.SPARhsn = None
        self.SPARhsa = None
        self.SPARcalibPow_M = None
        self.SPARcalibSNR_M = None
        self.SPARcalibPow_S = None
        self.SPARcalibSNR_S = None
        self.SPARrawGate1 = None
        self.SPARrawGate2 = None
        self.SPARraw = None
        self.SPARprc = None

        self.FHsize = 1180

    def FHread(self, fp):

        header = numpy.fromfile(fp, FILE_HEADER, 1)
        '''      numpy.fromfile(file, dtype, count, sep='')
            file : file or str
            Open file object or filename.
            
            dtype : data-type
            Data type of the returned array. For binary files, it is used to determine 
            the size and byte-order of the items in the file.
            
            count : int
            Number of items to read. -1 means all items (i.e., the complete file).
            
            sep : str
            Separator between items if file is a text file. Empty ("") separator means 
            the file should be treated as binary. Spaces (" ") in the separator match zero 
            or more whitespace characters. A separator consisting only of spaces must match 
            at least one whitespace.
            
    '''

        self.Hname = str(header['Hname'][0])
        self.Htime = str(header['Htime'][0])
        self.Hoper = str(header['Hoper'][0])
        self.Hplace = str(header['Hplace'][0])
        self.Hdescr = str(header['Hdescr'][0])
        self.Hdummy = str(header['Hdummy'][0])
        # 1024

        self.Msign = str(header['Msign'][0])
        self.MsizeData = header['MsizeData'][0]
        # 8

        self.PPARsign = str(header['PPARsign'][0])
        self.PPARsize = header['PPARsize'][0]
        self.PPARprf = header['PPARprf'][0]
        self.PPARpdr = header['PPARpdr'][0]
        self.PPARsft = header['PPARsft'][0]
        self.PPARavc = header['PPARavc'][0]
        self.PPARihp = header['PPARihp'][0]
        self.PPARchg = header['PPARchg'][0]
        self.PPARpol = header['PPARpol'][0]
        # Service DSP parameters
        # 36

        self.SPARatt = header['SPARatt'][0]
        self.SPARtx = header['SPARtx'][0]
        self.SPARaddGain0 = header['SPARaddGain0'][0]
        self.SPARaddGain1 = header['SPARaddGain1'][0]
        self.SPARwnd = header['SPARwnd'][0]
        self.SPARpos = header['SPARpos'][0]
        self.SPARadd = header['SPARadd'][0]
        self.SPARlen = header['SPARlen'][0]
        self.SPARcal = header['SPARcal'][0]
        self.SPARnos = header['SPARnos'][0]
        self.SPARof0 = header['SPARof0'][0]
        self.SPARof1 = header['SPARof1'][0]
        self.SPARswt = header['SPARswt'][0]
        self.SPARsum = header['SPARsum'][0]
        self.SPARosc = header['SPARosc'][0]
        self.SPARtst = header['SPARtst'][0]
        self.SPARcor = header['SPARcor'][0]
        self.SPARofs = header['SPARofs'][0]
        self.SPARhsn = header['SPARhsn'][0]
        self.SPARhsa = header['SPARhsa'][0]
        self.SPARcalibPow_M = header['SPARcalibPow_M'][0]
        self.SPARcalibSNR_M = header['SPARcalibSNR_M'][0]
        self.SPARcalibPow_S = header['SPARcalibPow_S'][0]
        self.SPARcalibSNR_S = header['SPARcalibSNR_S'][0]
        self.SPARrawGate1 = header['SPARrawGate1'][0]
        self.SPARrawGate2 = header['SPARrawGate2'][0]
        self.SPARraw = header['SPARraw'][0]
        self.SPARprc = header['SPARprc'][0]
        # 112
        # 1180
        # print 'Pointer fp header', fp.tell()
        # print ' '
        # print 'SPARrawGate'
        # print self.SPARrawGate2 - self.SPARrawGate1

        # print ' '
        # print 'Hname'
        # print self.Hname

        # print ' '
        # print 'Msign'
        # print self.Msign

    def write(self, fp):

        headerTuple = (self.Hname,
                       self.Htime,
                       self.Hoper,
                       self.Hplace,
                       self.Hdescr,
                       self.Hdummy)

        header = numpy.array(headerTuple, FILE_HEADER)
        #        numpy.array(object, dtype=None, copy=True, order=None, subok=False, ndmin=0)
        header.tofile(fp)
        ''' ndarray.tofile(fid, sep, format)    Write array to a file as text or binary (default).

        fid : file or str
        An open file object, or a string containing a filename.

        sep : str
        Separator between array items for text output. If "" (empty), a binary file is written, 
        equivalent to file.write(a.tobytes()).

        format : str
        Format string for text file output. Each entry in the array is formatted to text by 
        first converting it to the closest Python type, and then using "format" % item.

        '''

        return 1


SRVI_HEADER = numpy.dtype([
                         ('SignatureSRVI1', numpy.str_, 4),
                         ('SizeOfDataBlock1', '<i4'),
                         ('DataBlockTitleSRVI1', numpy.str_, 4),
                         ('SizeOfSRVI1', '<i4'), ])


class SRVIHeader(Header):
    def __init__(self,     SignatureSRVI1=0,     SizeOfDataBlock1=0,     DataBlockTitleSRVI1=0,    SizeOfSRVI1=0):

        self.SignatureSRVI1 = SignatureSRVI1
        self.SizeOfDataBlock1 = SizeOfDataBlock1
        self.DataBlockTitleSRVI1 = DataBlockTitleSRVI1
        self.SizeOfSRVI1 = SizeOfSRVI1

        self.SRVIHsize = 16

    def SRVIread(self, fp):

        header = numpy.fromfile(fp, SRVI_HEADER, 1)

        self.SignatureSRVI1 = str(header['SignatureSRVI1'][0])
        self.SizeOfDataBlock1 = header['SizeOfDataBlock1'][0]
        self.DataBlockTitleSRVI1 = str(header['DataBlockTitleSRVI1'][0])
        self.SizeOfSRVI1 = header['SizeOfSRVI1'][0]
        # 16
        print('Pointer fp SRVIheader', fp.tell())


SRVI_STRUCTURE = numpy.dtype([
                            ('frame_cnt', '<u4'),
                            ('time_t', '<u4'),   #
                            ('tpow', '<f4'),     #
                            ('npw1', '<f4'),     #
                            ('npw2', '<f4'),     #
                            ('cpw1', '<f4'),     #
                            ('pcw2', '<f4'),     #
                            ('ps_err', '<u4'),   #
                            ('te_err', '<u4'),   #
                            ('rc_err', '<u4'),   #
                            ('grs1', '<u4'),     #
                            ('grs2', '<u4'),     #
                            ('azipos', '<f4'),     #
                            ('azivel', '<f4'),     #
                            ('elvpos', '<f4'),     #
                            ('elvvel', '<f4'),     #
                            ('northAngle', '<f4'),
                            ('microsec', '<u4'),   #
                            ('azisetvel', '<f4'),  #
                            ('elvsetpos', '<f4'),  #
                            ('RadarConst', '<f4'), ])   #


class RecordHeader(Header):

    def __init__(self,     frame_cnt=0,  time_t=0,  tpow=0,   npw1=0,   npw2=0,
                 cpw1=0,   pcw2=0,       ps_err=0,   te_err=0,   rc_err=0,   grs1=0,
                 grs2=0,   azipos=0,   azivel=0,   elvpos=0,   elvvel=0,   northangle=0,
                 microsec=0,   azisetvel=0,   elvsetpos=0,   RadarConst=0, RecCounter=0, Off2StartNxtRec=0):

        self.frame_cnt = frame_cnt
        self.dwell = time_t
        self.tpow = tpow
        self.npw1 = npw1
        self.npw2 = npw2
        self.cpw1 = cpw1
        self.pcw2 = pcw2
        self.ps_err = ps_err
        self.te_err = te_err
        self.rc_err = rc_err
        self.grs1 = grs1
        self.grs2 = grs2
        self.azipos = azipos
        self.azivel = azivel
        self.elvpos = elvpos
        self.elvvel = elvvel
        self.northAngle = northangle
        self.microsec = microsec
        self.azisetvel = azisetvel
        self.elvsetpos = elvsetpos
        self.RadarConst = RadarConst
        self.RHsize = 84
        self.RecCounter = RecCounter
        self.Off2StartNxtRec = Off2StartNxtRec

    def RHread(self, fp):

        # startFp = open(fp,"rb") #The method tell() returns the current position of the file read/write pointer within the file.

        #OffRHeader= 1180 + self.RecCounter*(self.Off2StartNxtRec)
        #startFp.seek(OffRHeader, os.SEEK_SET)

        # print 'Posicion del bloque:        ',OffRHeader

        header = numpy.fromfile(fp, SRVI_STRUCTURE, 1)

        self.frame_cnt = header['frame_cnt'][0]
        self.time_t = header['time_t'][0]   #
        self.tpow = header['tpow'][0]     #
        self.npw1 = header['npw1'][0]     #
        self.npw2 = header['npw2'][0]     #
        self.cpw1 = header['cpw1'][0]     #
        self.pcw2 = header['pcw2'][0]     #
        self.ps_err = header['ps_err'][0]    #
        self.te_err = header['te_err'][0]    #
        self.rc_err = header['rc_err'][0]    #
        self.grs1 = header['grs1'][0]      #
        self.grs2 = header['grs2'][0]      #
        self.azipos = header['azipos'][0]     #
        self.azivel = header['azivel'][0]     #
        self.elvpos = header['elvpos'][0]     #
        self.elvvel = header['elvvel'][0]     #
        self.northAngle = header['northAngle'][0]    #
        self.microsec = header['microsec'][0]      #
        self.azisetvel = header['azisetvel'][0]     #
        self.elvsetpos = header['elvsetpos'][0]     #
        self.RadarConst = header['RadarConst'][0]    #
        # 84

        # print 'Pointer fp RECheader', fp.tell()

        #self.ipp= 0.5*(SPEED_OF_LIGHT/self.PRFhz)

        #self.RHsize = 180+20*self.nChannels
        #self.Datasize= self.nProfiles*self.nChannels*self.nHeights*2*4
        # print 'Datasize',self.Datasize
        #endFp = self.OffsetStartHeader + self.RecCounter*self.Off2StartNxtRec

        print('==============================================')

        print('==============================================')

        return 1


class MIRA35CReader (ProcessingUnit, FileHeaderMIRA35c, SRVIHeader, RecordHeader):

    path = None
    startDate = None
    endDate = None
    startTime = None
    endTime = None
    walk = None
    isConfig = False

    fileList = None

    # metadata
    TimeZone = None
    Interval = None
    heightList = None

    # data
    data = None
    utctime = None

    def __init__(self, **kwargs):

        # Eliminar de la base la herencia
        ProcessingUnit.__init__(self, **kwargs)
        self.PointerReader = 0
        self.FileHeaderFlag = False
        self.utc = None
        self.ext = ".zspca"
        self.optchar = "P"
        self.fpFile = None
        self.fp = None
        self.BlockCounter = 0
        self.dtype = None
        self.fileSizeByHeader = None
        self.filenameList = []
        self.fileSelector = 0
        self.Off2StartNxtRec = 0
        self.RecCounter = 0
        self.flagNoMoreFiles = 0
        self.data_spc = None
        # self.data_cspc=None
        self.data_output = None
        self.path = None
        self.OffsetStartHeader = 0
        self.Off2StartData = 0
        self.ipp = 0
        self.nFDTdataRecors = 0
        self.blocksize = 0
        self.dataOut = Spectra()
        self.profileIndex = 1  # Always
        self.dataOut.flagNoData = False
        self.dataOut.nRdPairs = 0        
        self.dataOut.data_spc = None        
        self.nextfileflag = True
        self.dataOut.RadarConst = 0
        self.dataOut.HSDV = []
        self.dataOut.NPW = []
        self.dataOut.COFA = []
        # self.dataOut.noise = 0

    def Files2Read(self, fp):
        ''' 
        Function that indicates the number of .fdt files that exist in the folder to be read.
        It also creates an organized list with the names of the files to read.
        '''
        # self.__checkPath()

        # Gets the list of files within the fp address
        ListaData = os.listdir(fp)
        # Sort the list of files from least to largest by names
        ListaData = sorted(ListaData)
        nFiles = 0  # File Counter
        FileList = []  # A list is created that will contain the .fdt files
        for IndexFile in ListaData:
            if '.zspca' in IndexFile and '.gz' not in IndexFile:
                FileList.append(IndexFile)
                nFiles += 1

        # print 'Files2Read'
        # print 'Existen '+str(nFiles)+' archivos .fdt'

        self.filenameList = FileList  # List of files from least to largest by names

    def run(self, **kwargs):
        '''
        This method will be the one that will initiate the data entry, will be called constantly.
        You should first verify that your Setup () is set up and then continue to acquire
        the data to be processed with getData ().
        '''
        if not self.isConfig:
            self.setup(**kwargs)
            self.isConfig = True

        self.getData()

    def setup(self, path=None,
              startDate=None,
              endDate=None,
              startTime=None,
              endTime=None,
              walk=True,
              timezone='utc',
              code=None,
              online=False,
              ReadMode=None, **kwargs):

        self.isConfig = True

        self.path = path
        self.startDate = startDate
        self.endDate = endDate
        self.startTime = startTime
        self.endTime = endTime
        self.walk = walk
        # self.ReadMode=int(ReadMode)

        pass

    def getData(self):
        '''
        Before starting this function, you should check that there is still an unread file,
        If there are still blocks to read or if the data block is empty.

        You should call the file "read".

        '''

        if self.flagNoMoreFiles:
            self.dataOut.flagNoData = True
            print('NoData se vuelve true')
            return 0

        self.fp = self.path
        self.Files2Read(self.fp)
        self.readFile(self.fp)

        self.dataOut.data_spc = self.dataOut_spc  # self.data_spc.copy()
        self.dataOut.RadarConst = self.RadarConst
        self.dataOut.data_output = self.data_output
        self.dataOut.noise = self.dataOut.getNoise()
        # print 'ACAAAAAA', self.dataOut.noise
        self.dataOut.data_spc = self.dataOut.data_spc + self.dataOut.noise
        self.dataOut.normFactor = 1
        # print 'self.dataOut.noise',self.dataOut.noise

        return self.dataOut.data_spc

    def readFile(self, fp):
        '''
        You must indicate if you are reading in Online or Offline mode and load the
        The parameters for this file reading mode.

        Then you must do 2 actions:

        1. Get the BLTR FileHeader.
        2. Start reading the first block.
        '''

        # The address of the folder is generated the name of the .fdt file that will be read
        print("File:    ", self.fileSelector + 1)

        if self.fileSelector < len(self.filenameList):

            self.fpFile = str(fp) + '/' + \
                str(self.filenameList[self.fileSelector])

            if self.nextfileflag == True:
                self.fp = open(self.fpFile, "rb")
                self.nextfileflag == False

            '''HERE STARTING THE FILE READING'''

            self.fheader = FileHeaderMIRA35c()
            self.fheader.FHread(self.fp)  # Bltr FileHeader Reading

            self.SPARrawGate1 = self.fheader.SPARrawGate1
            self.SPARrawGate2 = self.fheader.SPARrawGate2
            self.Num_Hei = self.SPARrawGate2 - self.SPARrawGate1
            self.Num_Bins = self.fheader.PPARsft
            self.dataOut.nFFTPoints = self.fheader.PPARsft

            self.Num_inCoh = self.fheader.PPARavc
            self.dataOut.PRF = self.fheader.PPARprf
            self.dataOut.frequency = 34.85 * 10**9
            self.Lambda = SPEED_OF_LIGHT / self.dataOut.frequency
            self.dataOut.ippSeconds = 1. / float(self.dataOut.PRF)

            pulse_width = self.fheader.PPARpdr * 10**-9
            self.__deltaHeigth = 0.5 * SPEED_OF_LIGHT * pulse_width

            self.data_spc = numpy.zeros((self.Num_Hei, self.Num_Bins, 2))
            self.dataOut.HSDV = numpy.zeros((self.Num_Hei, 2))

            self.Ze = numpy.zeros(self.Num_Hei)
            self.ETA = numpy.zeros(([2, self.Num_Hei]))

            self.readBlock()  # Block reading

        else:
            print('readFile FlagNoData becomes true')
            self.flagNoMoreFiles = True
            self.dataOut.flagNoData = True
            self.FileHeaderFlag == True
            return 0

    def readBlock(self):
        '''
        It should be checked if the block has data, if it is not passed to the next file.

        Then the following is done:

        1. Read the RecordHeader
        2. Fill the buffer with the current block number.

        '''

        if self.PointerReader > 1180:
            self.fp.seek(self.PointerReader, os.SEEK_SET)
            self.FirstPoint = self.PointerReader

        else:
            self.FirstPoint = 1180

        self.srviHeader = SRVIHeader()

        self.srviHeader.SRVIread(self.fp)  # Se obtiene la cabecera del SRVI

        self.blocksize = self.srviHeader.SizeOfDataBlock1  # Se obtiene el tamao del bloque

        if self.blocksize == 148:
            print('blocksize == 148 bug')
            jump = numpy.fromfile(self.fp, [('jump', numpy.str_, 140)], 1)

            # Se obtiene la cabecera del SRVI
            self.srviHeader.SRVIread(self.fp)

        if not self.srviHeader.SizeOfSRVI1:
            self.fileSelector += 1
            self.nextfileflag == True
            self.FileHeaderFlag == True

        self.recordheader = RecordHeader()
        self.recordheader.RHread(self.fp)
        self.RadarConst = self.recordheader.RadarConst
        dwell = self.recordheader.time_t
        npw1 = self.recordheader.npw1
        npw2 = self.recordheader.npw2

        self.dataOut.channelList = list(range(1))
        self.dataOut.nIncohInt = self.Num_inCoh
        self.dataOut.nProfiles = self.Num_Bins
        self.dataOut.nCohInt = 1
        self.dataOut.windowOfFilter = 1
        self.dataOut.utctime = dwell
        self.dataOut.timeZone = 0

        self.dataOut.outputInterval = self.dataOut.getTimeInterval()
        self.dataOut.heightList = self.SPARrawGate1 * self.__deltaHeigth + \
            numpy.array(list(range(self.Num_Hei))) * self.__deltaHeigth

        self.HSDVsign = numpy.fromfile(self.fp, [('HSDV', numpy.str_, 4)], 1)
        self.SizeHSDV = numpy.fromfile(self.fp, [('SizeHSDV', '<i4')], 1)
        self.HSDV_Co = numpy.fromfile(
            self.fp, [('HSDV_Co', '<f4')], self.Num_Hei)
        self.HSDV_Cx = numpy.fromfile(
            self.fp, [('HSDV_Cx', '<f4')], self.Num_Hei)

        self.COFAsign = numpy.fromfile(self.fp, [('COFA', numpy.str_, 4)], 1)
        self.SizeCOFA = numpy.fromfile(self.fp, [('SizeCOFA', '<i4')], 1)
        self.COFA_Co = numpy.fromfile(
            self.fp, [('COFA_Co', '<f4')], self.Num_Hei)
        self.COFA_Cx = numpy.fromfile(
            self.fp, [('COFA_Cx', '<f4')], self.Num_Hei)

        self.ZSPCsign = numpy.fromfile(
            self.fp, [('ZSPCsign', numpy.str_, 4)], 1)
        self.SizeZSPC = numpy.fromfile(self.fp, [('SizeZSPC', '<i4')], 1)

        self.dataOut.HSDV[0] = self.HSDV_Co[:][0]
        self.dataOut.HSDV[1] = self.HSDV_Cx[:][0]

        for irg in range(self.Num_Hei):
            # Number of spectral sub pieces containing significant power
            nspc = numpy.fromfile(self.fp, [('nspc', 'int16')], 1)[0][0]

            for k in range(nspc):
                # Index of the spectral bin where the piece is beginning
                binIndex = numpy.fromfile(
                    self.fp, [('binIndex', 'int16')], 1)[0][0]
                nbins = numpy.fromfile(self.fp, [('nbins', 'int16')], 1)[
                    0][0]  # Number of bins of the piece

                # Co_Channel
                jbin = numpy.fromfile(self.fp, [('jbin', 'uint16')], nbins)[
                    0][0]  # Spectrum piece to be normaliced
                jmax = numpy.fromfile(self.fp, [('jmax', 'float32')], 1)[
                    0][0]  # Maximun piece to be normaliced

                self.data_spc[irg, binIndex:binIndex + nbins, 0] = self.data_spc[irg,
                                                                                 binIndex:binIndex + nbins, 0] + jbin / 65530. * jmax

                # Cx_Channel
                jbin = numpy.fromfile(
                    self.fp, [('jbin', 'uint16')], nbins)[0][0]
                jmax = numpy.fromfile(self.fp, [('jmax', 'float32')], 1)[0][0]

                self.data_spc[irg, binIndex:binIndex + nbins, 1] = self.data_spc[irg,
                                                                                 binIndex:binIndex + nbins, 1] + jbin / 65530. * jmax

        for bin in range(self.Num_Bins):

            self.data_spc[:, bin, 0] = self.data_spc[:,
                                                     bin, 0] - self.dataOut.HSDV[:, 0]

            self.data_spc[:, bin, 1] = self.data_spc[:,
                                                     bin, 1] - self.dataOut.HSDV[:, 1]

        numpy.set_printoptions(threshold='nan')

        self.data_spc = numpy.where(self.data_spc > 0., self.data_spc, 0)

        self.dataOut.COFA = numpy.array([self.COFA_Co, self.COFA_Cx])

        print(' ')
        print('SPC', numpy.shape(self.dataOut.data_spc))
        # print 'SPC',self.dataOut.data_spc

        noinor1 = 713031680
        noinor2 = 30

        npw1 = 1  # 0**(npw1/10) * noinor1 * noinor2
        npw2 = 1  # 0**(npw2/10) * noinor1 * noinor2
        self.dataOut.NPW = numpy.array([npw1, npw2])

        print(' ')

        self.data_spc = numpy.transpose(self.data_spc, (2, 1, 0))
        self.data_spc = numpy.fft.fftshift(self.data_spc, axes=1)

        self.data_spc = numpy.fliplr(self.data_spc)

        self.data_spc = numpy.where(self.data_spc > 0., self.data_spc, 0)
        self.dataOut_spc = numpy.ones([1, self.Num_Bins, self.Num_Hei])
        self.dataOut_spc[0, :, :] = self.data_spc[0, :, :]
        # print 'SHAPE', self.dataOut_spc.shape
        # For nyquist correction:
        # fix = 20 # ~3m/s
        #shift = self.Num_Bins/2 + fix
        #self.data_spc = numpy.array([ self.data_spc[: , self.Num_Bins-shift+1: , :] , self.data_spc[: , 0:self.Num_Bins-shift , :]])

        '''Block Reading, the Block Data is received and Reshape is used to give it
        shape.
        '''

        self.PointerReader = self.fp.tell()


'''
@author: Daniel Suarez
'''
import os
import glob
import ftplib

try:
    import paramiko
    import scp
except:
    pass

import time

import threading
Thread = threading.Thread

# try:
#     from gevent import sleep
# except:
from time import sleep

from schainpy.model.proc.jroproc_base import ProcessingUnit, Operation

class Remote(Thread):
    """
    Remote is a parent class used to define the behaviour of FTP and SSH class. These clases are
    used to upload or download files remotely.

    Non-standard Python modules used:
        None

    Written by:
        "Miguel Urco":mailto:miguel.urco@jro.igp.gob.pe Jun. 03, 2015
    """

    server = None
    username = None
    password = None
    remotefolder = None

    period = 60
    fileList = []
    bussy = False

    def __init__(self, server, username, password, remotefolder, period=60):

        Thread.__init__(self)

        self.setDaemon(True)

        self.status = 0

        self.__server = server
        self.__username = username
        self.__password = password
        self.__remotefolder = remotefolder

        self.period = period

        self.fileList = []
        self.bussy = False

        self.stopFlag = False

        print("[Remote Server] Opening server: %s" %self.__server)
        if self.open(self.__server, self.__username, self.__password, self.__remotefolder):
            print("[Remote Server] %s server was opened successfully" %self.__server)

        self.close()

        self.mutex = threading.Lock()

    def stop(self):

        self.stopFlag = True
        self.join(10)

    def open(self):
        """
        Connect to server and create a connection class (FTP or SSH) to remote server.
        """
        raise NotImplementedError("Implement this method in child class")

    def close(self):
        """
        Close connection to server
        """
        raise NotImplementedError("Implement this method in child class")

    def mkdir(self, remotefolder):
        """
        Create a folder remotely
        """
        raise NotImplementedError("Implement this method in child class")

    def cd(self, remotefolder):
        """
        Change working directory in remote server
        """
        raise NotImplementedError("Implement this method in child class")

    def download(self, filename, localfolder=None):
        """
        Download a file from server to local host
        """
        raise NotImplementedError("Implement this method in child class")

    def sendFile(self, fullfilename):
        """
        sendFile method is used to upload a local file to the current directory in remote server

        Inputs:
            fullfilename        - full path name of local file to store in remote directory

        Returns:
            0 in error case else 1
        """
        raise NotImplementedError("Implement this method in child class")

    def upload(self, fullfilename, remotefolder=None):
        """
        upload method is used to upload a local file to remote directory. This method changes
        working directory before sending a file.

        Inputs:
            fullfilename        - full path name of local file to store in remote directory

            remotefolder    - remote directory

        Returns:
            0 in error case else 1
        """
        print("[Remote Server] Uploading %s to %s:%s" %(fullfilename, self.server, self.remotefolder))

        if not self.status:
            return 0

        if remotefolder == None:
            remotefolder = self.remotefolder

        if not self.cd(remotefolder):
            return 0

        if not self.sendFile(fullfilename):
            print("[Remote Server] Error uploading file %s" %fullfilename)
            return 0

        print("[Remote Server] upload finished successfully")

        return 1

    def delete(self, filename):
        """
        Remove a file from remote server
        """
        pass

    def updateFileList(self, fileList):
        """
        Remove a file from remote server
        """

        if fileList == self.fileList:
            return 0

        self.mutex.acquire()
        #         init = time.time()
        #         
        #         while(self.bussy):
        #             sleep(0.1)
        #             if time.time() - init > 2*self.period:
        #                 return 0
            
        self.fileList = fileList
        self.mutex.release()
        return 1

    def run(self):

        if not self.status:
            print("Finishing FTP service")
            return

        if not self.cd(self.remotefolder):
            raise ValueError("Could not access to the new remote directory: %s" %self.remotefolder)

        while True:

            for i in range(self.period):
                if self.stopFlag:
                    break
                sleep(1)

            if self.stopFlag:
                break
                
            #   self.bussy = True
            self.mutex.acquire()

            print("[Remote Server] Opening %s" %self.__server)
            if not self.open(self.__server, self.__username, self.__password, self.__remotefolder):
                self.mutex.release()
                continue

            for thisFile in self.fileList:
                self.upload(thisFile, self.remotefolder)

            print("[Remote Server] Closing %s" %self.__server)
            self.close()

            self.mutex.release()
            # self.bussy = False

        print("[Remote Server] Thread stopped successfully")

class FTPClient(Remote):

    __ftpClientObj = None

    def __init__(self, server, username, password, remotefolder, period=60):
        """
        """
        Remote.__init__(self, server, username, password, remotefolder, period)

    def open(self, server, username, password, remotefolder):

        """
        This method is used to set FTP parameters and establish a connection to remote server

        Inputs:
            server    - remote server IP Address

            username    - remote server Username

            password    - remote server password

            remotefolder    - remote server current working directory

        Return:
            Boolean     -     Returns 1 if a connection has been established, 0 otherwise

        Affects:
            self.status        - in case of error or fail connection this parameter is set to 0 else 1

        """

        if server == None:
            raise ValueError("FTP server should be defined")

        if username == None:
            raise ValueError("FTP username should be defined")

        if password == None:
            raise ValueError("FTP password should be defined")

        if remotefolder == None:
            raise ValueError("FTP remote folder should be defined")

        try:
            ftpClientObj = ftplib.FTP(server)
        except ftplib.all_errors as e:
            print("[FTP Server]: FTP server connection fail: %s" %server)
            print("[FTP Server]:", e)
            self.status = 0
            return 0

        try:
            ftpClientObj.login(username, password)
        except ftplib.all_errors:
            print("[FTP Server]: FTP username or password are incorrect")
            self.status = 0
            return 0

        if remotefolder == None:
            remotefolder = ftpClientObj.pwd()
        else:
            try:
                ftpClientObj.cwd(remotefolder)
            except ftplib.all_errors:
                print("[FTP Server]: FTP remote folder is invalid: %s" %remotefolder)
                remotefolder = ftpClientObj.pwd()

        self.server = server
        self.username = username
        self.password = password
        self.remotefolder = remotefolder
        self.__ftpClientObj = ftpClientObj
        self.status = 1

        return 1

    def close(self):
        """
        Close connection to remote server
        """
        if not self.status:
            return 0

        self.__ftpClientObj.close()

    def mkdir(self, remotefolder):
        """
        mkdir is used to make a new directory in remote server

        Input:
            remotefolder    - directory name

        Return:
            0 in error case else 1
        """
        if not self.status:
            return 0

        try:
            self.__ftpClientObj.mkd(dirname)
        except ftplib.all_errors:
            print("[FTP Server]: Error creating remote folder: %s" %remotefolder)
            return 0

        return 1

    def cd(self, remotefolder):
        """
        cd is used to change remote working directory on server

        Input:
            remotefolder    - current working directory

        Affects:
            self.remotefolder

        Return:
            0 in case of error else 1
        """
        if not self.status:
            return 0

        if remotefolder == self.remotefolder:
            return 1

        try:
            self.__ftpClientObj.cwd(remotefolder)
        except ftplib.all_errors:
            print('[FTP Server]: Error changing to %s' %remotefolder)
            print('[FTP Server]: Trying to create remote folder')

            if not self.mkdir(remotefolder):
                print('[FTP Server]: Remote folder could not be created')
                return 0

            try:
                self.__ftpClientObj.cwd(remotefolder)
            except ftplib.all_errors:
                return 0

        self.remotefolder = remotefolder

        return 1

    def sendFile(self, fullfilename):

        if not self.status:
            return 0

        fp = open(fullfilename, 'rb')

        filename = os.path.basename(fullfilename)

        command = "STOR %s" %filename

        try:
            self.__ftpClientObj.storbinary(command, fp)
        except ftplib.all_errors as e:
            print("[FTP Server]:", e)
            return 0

        try:
            self.__ftpClientObj.sendcmd('SITE CHMOD 755 ' + filename)
        except ftplib.all_errors as e:
            print("[FTP Server]:", e)

        fp.close()

        return 1

class SSHClient(Remote):

    __sshClientObj = None
    __scpClientObj = None

    def __init__(self, server, username, password, remotefolder, period=60):
        """
        """
        Remote.__init__(self, server, username, password, remotefolder, period)

    def open(self, server, username, password, remotefolder, port=22):

        """
            This method is used to set SSH parameters and establish a connection to a remote server
            
            Inputs:
                server    - remote server IP Address 
                
                username    - remote server Username 
                
                password    - remote server password
                
                remotefolder    - remote server current working directory
            
            Return: void
            
            Affects: 
                self.status        - in case of error or fail connection this parameter is set to 0 else 1

        """
        import socket

        if server == None:
            raise ValueError("SSH server should be defined")

        if username == None:
            raise ValueError("SSH username should be defined")

        if password == None:
            raise ValueError("SSH password should be defined")

        if remotefolder == None:
            raise ValueError("SSH remote folder should be defined")

        sshClientObj = paramiko.SSHClient()

        sshClientObj.load_system_host_keys()
        sshClientObj.set_missing_host_key_policy(paramiko.WarningPolicy())

        self.status = 0
        try:
            sshClientObj.connect(server, username=username, password=password, port=port)
        except paramiko.AuthenticationException as e:
    #             print "SSH username or password are incorrect: %s"
            print("[SSH Server]:", e)
            return 0
        except SSHException as e:
            print("[SSH Server]:", e)
            return 0
        except socket.error:
            self.status = 0
            print("[SSH Server]:", e)
            return 0

        self.status = 1
        scpClientObj = scp.SCPClient(sshClientObj.get_transport(), socket_timeout=30)

        if remotefolder == None:
            remotefolder = self.pwd()

        self.server = server
        self.username = username
        self.password = password
        self.__sshClientObj = sshClientObj
        self.__scpClientObj = scpClientObj
        self.status = 1

        if not self.cd(remotefolder):
            raise ValueError("[SSH Server]: Could not access to remote folder: %s" %remotefolder)
            return 0

        self.remotefolder = remotefolder

        return 1

    def close(self):
        """
            Close connection to remote server
        """
        if not self.status:
            return 0

        self.__scpClientObj.close()
        self.__sshClientObj.close()

    def __execute(self, command):
        """
            __execute a command on remote server
            
            Input:
                command    - Exmaple 'ls -l'
            
            Return:
                0 in error case else 1
        """
        if not self.status:
            return 0

        stdin, stdout, stderr = self.__sshClientObj.exec_command(command)

        result = stderr.readlines()
        if len(result) > 1:
            return 0

        result = stdout.readlines()
        if len(result) > 1:
            return result[0][:-1]

        return 1

    def mkdir(self, remotefolder):
        """
            mkdir is used to make a new directory in remote server
            
            Input:
                remotefolder    - directory name
            
            Return:
                0 in error case else 1
        """

        command = 'mkdir %s' %remotefolder

        return self.__execute(command)

    def pwd(self):

        command = 'pwd'

        return self.__execute(command)

    def cd(self, remotefolder):
        """
            cd is used to change remote working directory on server
            
            Input:
                remotefolder    - current working directory
                
            Affects:
                self.remotefolder
            
            Return: 
                0 in case of error else 1
        """
        if not self.status:
            return 0

        if remotefolder == self.remotefolder:
            return 1

        chk_command = "cd %s; pwd" %remotefolder
        mkdir_command = "mkdir %s" %remotefolder

        if not self.__execute(chk_command):
            if not self.__execute(mkdir_command):
                self.remotefolder = None
                return 0

        self.remotefolder = remotefolder

        return 1

    def sendFile(self, fullfilename):

        if not self.status:
            return 0

        try:
            self.__scpClientObj.put(fullfilename, remote_path=self.remotefolder)
        except scp.ScpError as e:
            print("[SSH Server]", str(e))
            return 0

        remotefile = os.path.join(self.remotefolder, os.path.split(fullfilename)[-1])
        command = 'chmod 775 %s' %remotefile

        return self.__execute(command)

class SendToServer(ProcessingUnit):

    def __init__(self, **kwargs):

        ProcessingUnit.__init__(self, **kwargs)

        self.isConfig = False
        self.clientObj = None        
    
    def setup(self, server, username, password, remotefolder, localfolder, ext='.png', period=60, protocol='ftp', **kwargs):

        self.clientObj = None
        self.localfolder = localfolder
        self.ext = ext
        self.period = period

        if str.lower(protocol) == 'ftp':
            self.clientObj = FTPClient(server, username, password, remotefolder, period)

        if str.lower(protocol) == 'ssh':
            self.clientObj = SSHClient(server, username, password, remotefolder, period)

        if not self.clientObj:
            raise ValueError("%s has been chosen as remote access protocol but it is not valid" %protocol)

        self.clientObj.start()

    def findFiles(self):

        if not type(self.localfolder) == list:
            folderList = [self.localfolder]
        else:
            folderList = self.localfolder

        #Remove duplicate items
        folderList = list(set(folderList))

        fullfilenameList = []

        for thisFolder in folderList:

            print("[Remote Server]: Searching files on %s" %thisFolder)

            filenameList = glob.glob1(thisFolder, '*%s' %self.ext)

            if len(filenameList) < 1:

                continue

            for thisFile in filenameList:
                fullfilename = os.path.join(thisFolder, thisFile)

                if fullfilename in fullfilenameList:
                    continue

                #Only files modified in the last 30 minutes are considered
                if os.path.getmtime(fullfilename) < time.time() - 30*60:
                    continue

                fullfilenameList.append(fullfilename)

        return fullfilenameList

    def run(self, **kwargs):
        if not self.isConfig:
            self.init = time.time()
            self.setup(**kwargs)
            self.isConfig = True
        
        if not self.clientObj.is_alive():
            print("[Remote Server]: Restarting connection ")
            self.setup(**kwargs)
        
        if time.time() - self.init >= self.period:
            fullfilenameList = self.findFiles()

            if self.clientObj.updateFileList(fullfilenameList):
                print("[Remote Server]: Sending the next files ", str(fullfilenameList))
            self.init = time.time()

    def close(self):
        print("[Remote Server] Stopping thread")
        self.clientObj.stop()


class FTP(object):
    """
    Ftp is a public class used to define custom File Transfer Protocol from "ftplib" python module

    Non-standard Python modules used: None

    Written by "Daniel Suarez":mailto:daniel.suarez@jro.igp.gob.pe  Oct. 26, 2010
    """

    def __init__(self,server = None, username=None, password=None, remotefolder=None):
        """
        This method is used to setting parameters for FTP and establishing connection to remote server

        Inputs:
            server    - remote server IP Address

            username    - remote server Username

            password    - remote server password

            remotefolder    - remote server current working directory

        Return: void

        Affects:
            self.status    - in Error Case or Connection Failed this parameter is set to 1 else 0

            self.folderList    - sub-folder list of remote folder

            self.fileList    - file list of remote folder


        """

        if ((server == None) and (username==None) and (password==None) and (remotefolder==None)):
            server, username, password, remotefolder = self.parmsByDefault()

        self.server = server
        self.username = username
        self.password = password
        self.remotefolder = remotefolder
        self.file = None
        self.ftp = None
        self.status = 0

        try:
            self.ftp = ftplib.FTP(self.server)
            self.ftp.login(self.username,self.password)
            self.ftp.cwd(self.remotefolder)            
            #   print 'Connect to FTP Server: Successfully'
        
        except ftplib.all_errors:
            print('Error FTP Service')
            self.status = 1
            return



        self.dirList = []

        try:
            self.dirList = self.ftp.nlst()

        except ftplib.error_perm as resp:
            if str(resp) == "550 No files found":
                    print("no files in this directory")
                    self.status = 1
                    return

        except ftplib.all_errors:
            print('Error Displaying Dir-Files')
            self.status = 1
            return

        self.fileList = []
        self.folderList = []
        #only for test
        for f in self.dirList:
            name, ext = os.path.splitext(f)
            if ext != '':
                self.fileList.append(f)
    #                print 'filename: %s - size: %d'%(f,self.ftp.size(f))

    def parmsByDefault(self):
        server = 'jro-app.igp.gob.pe'
        username = 'wmaster'
        password = 'mst2010vhf'
        remotefolder = '/home/wmaster/graficos'

        return server, username, password, remotefolder


    def mkd(self,dirname):
        """
        mkd is used to make directory in remote server

        Input:
            dirname    - directory name

        Return:
            1 in error case else 0
        """
        try:
            self.ftp.mkd(dirname)
        except:
            print('Error creating remote folder:%s'%dirname)
            return 1

        return 0


    def delete(self,filename):
        """
        delete is used to delete file in current working directory of remote server

        Input:
            filename    - filename to delete in remote folder

        Return:
            1 in error case else 0
        """

        try:
            self.ftp.delete(filename)
        except:
            print('Error deleting remote file:%s'%filename)
            return 1

        return 0

    def download(self,filename,localfolder):
        """
        download is used to downloading file from remote folder into local folder

        Inputs:
            filename    - filename to donwload

            localfolder    - directory local to store filename

        Returns:
            self.status    - 1 in error case else 0
        """

        self.status = 0


        if not(filename in self.fileList):
            print('filename:%s not exists'%filename)
            self.status = 1
            return self.status

        newfilename = os.path.join(localfolder,filename)

        self.file = open(newfilename, 'wb')

        try:
            print('Download: ' + filename)
            self.ftp.retrbinary('RETR ' + filename, self.__handleDownload)
            print('Download Complete')
        except ftplib.all_errors:
            print('Error Downloading ' + filename)
            self.status = 1
            return self.status

        self.file.close()

        return self.status


    def __handleDownload(self,block):
        """
        __handleDownload is used to handle writing file
        """
        self.file.write(block)


    def upload(self,filename,remotefolder=None):
        """
        upload is used to uploading local file to remote directory

        Inputs:
            filename    - full path name of local file to store in remote directory

            remotefolder    - remote directory

        Returns:
            self.status    - 1 in error case else 0
        """

        if remotefolder == None:
            remotefolder = self.remotefolder

        self.status = 0

        try:
            self.ftp.cwd(remotefolder)

            self.file = open(filename, 'rb')

            (head, tail) = os.path.split(filename)

            command = "STOR " + tail

            print('Uploading: ' + tail)
            self.ftp.storbinary(command, self.file)
            print('Upload Completed')

        except ftplib.all_errors:
            print('Error Uploading ' + tail)
            self.status = 1
            return self.status

        self.file.close()

        #back to initial directory in __init__()
        self.ftp.cwd(self.remotefolder)

        return self.status


    def dir(self,remotefolder):
        """
        dir is used to change working directory of remote server and get folder and file list

        Input:
            remotefolder    - current working directory

        Affects:
            self.fileList    - file list of working directory

        Return:
            infoList    - list with filenames and size of file in bytes

            self.folderList    -    folder list
        """

        self.remotefolder = remotefolder
        print('Change to ' + self.remotefolder)
        try:
            self.ftp.cwd(remotefolder)
        except ftplib.all_errors:
            print('Error Change to ' + self.remotefolder)
            infoList = None
            self.folderList = None
            return infoList,self.folderList

        self.dirList = []

        try:
            self.dirList = self.ftp.nlst()

        except ftplib.error_perm as resp:
            if str(resp) == "550 No files found":
                    print("no files in this directory")
                    infoList = None
                    self.folderList = None
                    return infoList,self.folderList
        except ftplib.all_errors:
            print('Error Displaying Dir-Files')
            infoList = None
            self.folderList = None
            return infoList,self.folderList

        infoList = []
        self.fileList = []
        self.folderList = []
        for f in self.dirList:
            name,ext = os.path.splitext(f)
            if ext != '':
                self.fileList.append(f)
                value = (f,self.ftp.size(f))
                infoList.append(value)

            if ext == '':
                self.folderList.append(f)

        return infoList,self.folderList


    def close(self):
        """
        close is used to close and end FTP connection

        Inputs: None

        Return: void

        """
        self.ftp.close()

class SendByFTP(Operation):

    def __init__(self, **kwargs):
        Operation.__init__(self, **kwargs)
        self.status = 1
        self.counter = 0

    def error_print(self, ValueError):

        print(ValueError, 'Error FTP')
        print("don't worry the program is running...")

    def worker_ftp(self, server, username, password, remotefolder, filenameList):

        self.ftpClientObj = FTP(server, username, password, remotefolder)
        for filename in filenameList:
            self.ftpClientObj.upload(filename)
        self.ftpClientObj.close()

    def ftp_thread(self, server, username, password, remotefolder):
        if not(self.status):
            return

        import multiprocessing

        p = multiprocessing.Process(target=self.worker_ftp, args=(server, username, password, remotefolder, self.filenameList,))
        p.start()

        p.join(3)

        if p.is_alive():
            p.terminate()
            p.join()
            print('killing ftp process...')
            self.status = 0
            return

        self.status = 1
        return

    def filterByExt(self, ext, localfolder):
        fnameList = glob.glob1(localfolder,ext)
        self.filenameList = [os.path.join(localfolder,x) for x in fnameList]

        if len(self.filenameList) == 0:
            self.status = 0

    def run(self, dataOut, ext, localfolder, remotefolder, server, username, password, period=1):

        self.counter += 1
        if self.counter >= period:
            self.filterByExt(ext, localfolder)

            self.ftp_thread(server, username, password, remotefolder)

            self.counter = 0

        self.status = 1


'''
Created on Jan 15, 2018

@author: Juan C. Espinoza
'''

import os
import sys
import time
import glob
import datetime
import tarfile

import numpy

from .utils import folder_in_range

from schainpy.model.io.jroIO_base import JRODataReader
from schainpy.model.proc.jroproc_base import ProcessingUnit, Operation
from schainpy.model.data.jrodata import Parameters
from schainpy.utils import log

try:
    from netCDF4 import Dataset
except:
    pass

UT1970 = datetime.datetime(1970, 1, 1) - datetime.timedelta(seconds=time.timezone)


class PXReader(JRODataReader, ProcessingUnit):

    def __init__(self, **kwargs):

        ProcessingUnit.__init__(self, **kwargs)

        self.dataOut = Parameters()    
        self.counter_records = 0
        self.nrecords = None
        self.flagNoMoreFiles = 0
        self.isConfig = False        
        self.filename = None        
        self.intervals = set()
        self.ext = ('.nc', '.tgz')
        self.online_mode = False
        
    def setup(self,
              path=None,
              startDate=None,
              endDate=None,
              format=None,
              startTime=datetime.time(0, 0, 0),
              endTime=datetime.time(23, 59, 59),
              walk=False,
              **kwargs):
                
        self.path = path
        self.startDate = startDate
        self.endDate = endDate
        self.startTime = startTime
        self.endTime = endTime
        self.datatime = datetime.datetime(1900,1,1)
        self.walk = walk
        self.nTries = kwargs.get('nTries', 10)
        self.online = kwargs.get('online', False)
        self.delay = kwargs.get('delay', 60)
        self.ele = kwargs.get('ext', '')

        if self.path is None:
            raise ValueError('The path is not valid')

        self.search_files(path, startDate, endDate, startTime, endTime, walk)
        self.cursor = 0
        self.counter_records = 0

        if not self.files:
            raise  Warning('There is no files matching these date in the folder: {}. \n Check startDate and endDate'.format(path))
        
    def search_files(self, path, startDate, endDate, startTime, endTime, walk):
        '''
         Searching for NCDF files in path
         Creating a list of files to procces included in [startDate,endDate]
         
         Input: 
             path - Path to find files             
        '''    

        log.log('Searching files {} in {} '.format(self.ext, path), 'PXReader')
        if walk:
            paths = [os.path.join(path, p) for p in os.listdir(path) if os.path.isdir(os.path.join(path, p))]
            paths.sort()
        else:
            paths = [path]

        fileList0 = []
        
        for subpath in paths:
            if not folder_in_range(subpath.split('/')[-1], startDate, endDate, '%Y%m%d'):
                continue
            fileList0 += [os.path.join(subpath, s) for s in glob.glob1(subpath, '*') if os.path.splitext(s)[-1] in self.ext and '{}'.format(self.ele) in s]
        
        fileList0.sort()
        if self.online:
            fileList0 = fileList0[-1:]

        self.files = {}

        startDate = startDate - datetime.timedelta(1)
        endDate = endDate + datetime.timedelta(1)

        for fullname in fileList0:
            thisFile = fullname.split('/')[-1]
            year = thisFile[3:7]
            if not year.isdigit():
                continue

            month = thisFile[7:9]
            if not month.isdigit():
                continue

            day = thisFile[9:11]
            if not day.isdigit():
                continue
            
            year, month, day = int(year), int(month), int(day)
            dateFile = datetime.date(year, month, day)
            timeFile = datetime.time(int(thisFile[12:14]), int(thisFile[14:16]), int(thisFile[16:18]))
            
            if (startDate > dateFile) or (endDate < dateFile):
                continue
            
            dt = datetime.datetime.combine(dateFile, timeFile)
            if dt not in self.files:
                self.files[dt] = []
            self.files[dt].append(fullname)

        self.dates = list(self.files.keys())
        self.dates.sort()

        return

    def search_files_online(self):
        '''
         Searching for NCDF files in online mode path
         Creating a list of files to procces included in [startDate,endDate]
         
         Input: 
             path - Path to find files             
        '''    

        self.files = {}

        for n in range(self.nTries):

            if self.walk:
                paths = [os.path.join(self.path, p) for p in os.listdir(self.path) if os.path.isdir(os.path.join(self.path, p))]
                paths.sort()
                path = paths[-1]
            else:
                path = self.path

            new_files = [os.path.join(path, s) for s in glob.glob1(path, '*') if os.path.splitext(s)[-1] in self.ext and '{}'.format(self.ele) in s]            
            new_files.sort()

            for fullname in new_files:
                thisFile = fullname.split('/')[-1]
                year = thisFile[3:7]
                if not year.isdigit():
                    continue

                month = thisFile[7:9]
                if not month.isdigit():
                    continue

                day = thisFile[9:11]
                if not day.isdigit():
                    continue
                
                year, month, day = int(year), int(month), int(day)
                dateFile = datetime.date(year, month, day)
                timeFile = datetime.time(int(thisFile[12:14]), int(thisFile[14:16]), int(thisFile[16:18]))
                
                dt = datetime.datetime.combine(dateFile, timeFile)
                
                if self.dt >= dt:
                    continue
                
                if dt not in self.files:
                    self.dt = dt
                    self.files[dt] = []
                
                self.files[dt].append(fullname)
                break

            if self.files:
                break
            else:
                log.warning('Waiting {} seconds for the next file, try {} ...'.format(self.delay, n + 1), 'PXReader')
                time.sleep(self.delay)

        if not self.files:
            return 0

        self.dates = list(self.files.keys())
        self.dates.sort()
        self.cursor = 0

        return 1

    def parseFile(self):
        '''
        '''

        header = {}
        
        for attr in self.fp.ncattrs():
            header[str(attr)] = getattr(self.fp, attr)

        self.header.append(header)

        self.data[header['TypeName']] = numpy.array(self.fp.variables[header['TypeName']])

    def setNextFile(self):
        '''
        Open next files for the current datetime
        '''

        cursor = self.cursor
        if not self.online_mode:
            if cursor == len(self.dates):
                if self.online:
                    cursor = 0
                    self.dt = self.dates[cursor]
                    self.online_mode = True
                    if not self.search_files_online():
                        log.success('No more files', 'PXReader')
                        return 0
                else:
                    log.success('No more files', 'PXReader')
                    self.flagNoMoreFiles = 1
                    return 0
        else:
            if not self.search_files_online():
                return 0
            cursor = self.cursor

        self.data = {}
        self.header = []

        for fullname in self.files[self.dates[cursor]]:

            log.log('Opening: {}'.format(fullname), 'PXReader')

            if os.path.splitext(fullname)[-1] == '.tgz':
                tar = tarfile.open(fullname, 'r:gz')
                tar.extractall('/tmp')
                files = [os.path.join('/tmp', member.name) for member in tar.getmembers()]
            else:
                files = [fullname]
            
            for filename in files:
                if self.filename is not None:
                    self.fp.close()

                self.filename = filename
                self.filedate = self.dates[cursor]
                self.fp = Dataset(self.filename, 'r')
                self.parseFile()

        self.counter_records += 1
        self.cursor += 1
        return 1

    def readNextFile(self):

        while True:
            self.flagDiscontinuousBlock = 0
            if not self.setNextFile():                    
                return 0

            self.datatime = datetime.datetime.utcfromtimestamp(self.header[0]['Time'])
            
            if self.online:
                break

            if (self.datatime < datetime.datetime.combine(self.startDate, self.startTime)) or \
               (self.datatime > datetime.datetime.combine(self.endDate, self.endTime)):
                log.warning(
                    'Reading Record No. {}/{} -> {} [Skipping]'.format(
                        self.counter_records,
                        self.nrecords,
                        self.datatime.ctime()),
                    'PXReader')
                continue
            break

        log.log(
            'Reading Record No. {}/{} -> {}'.format(
                self.counter_records,
                self.nrecords,
                self.datatime.ctime()),
            'PXReader')

        return 1


    def set_output(self):
        '''
        Storing data from buffer to dataOut object
        '''

        self.data['Elevation'] = numpy.array(self.fp.variables['Elevation'])
        self.data['Azimuth'] = numpy.array(self.fp.variables['Azimuth'])
        self.dataOut.range = numpy.array(self.fp.variables['GateWidth'])
        self.dataOut.data = self.data
        self.dataOut.units = [h['Unit-value'] for h in self.header]
        self.dataOut.parameters = [h['TypeName'] for h in self.header]
        self.dataOut.missing = self.header[0]['MissingData']
        self.dataOut.max_range = self.header[0]['MaximumRange-value']
        self.dataOut.elevation = self.header[0]['Elevation']
        self.dataOut.azimuth = self.header[0]['Azimuth']
        self.dataOut.latitude = self.header[0]['Latitude']
        self.dataOut.longitude = self.header[0]['Longitude']
        self.dataOut.utctime = self.header[0]['Time']
        self.dataOut.utctimeInit = self.dataOut.utctime
        self.dataOut.useLocalTime = True
        self.dataOut.flagNoData = False
        self.dataOut.flagDiscontinuousBlock = self.flagDiscontinuousBlock

        log.log('Parameters found: {}'.format(','.join(self.dataOut.parameters)),
                    'PXReader')

    def getData(self):
        '''
        Storing data from databuffer to dataOut object
        '''
        if self.flagNoMoreFiles:
            self.dataOut.flagNoData = True
            log.error('No file left to process', 'PXReader')
            return 0

        if not self.readNextFile():
            self.dataOut.flagNoData = True
            return 0

        self.set_output()

        return 1



import numpy

from .jroproc_base import ProcessingUnit, Operation
from schainpy.model.data.jrodata import Spectra
from schainpy.model.data.jrodata import hildebrand_sekhon

class SpectraLagsProc(ProcessingUnit):

    def __init__(self, **kwargs):

        ProcessingUnit.__init__(self, **kwargs)

        self.__input_buffer = None
        self.firstdatatime = None
        self.profIndex = 0
        self.dataOut = Spectra()
        self.id_min = None
        self.id_max = None
        self.__codeIndex = 0

        self.__lags_buffer = None

    def __updateSpecFromVoltage(self):

        self.dataOut.plotting = "spectra_lags"
        self.dataOut.timeZone = self.dataIn.timeZone
        self.dataOut.dstFlag = self.dataIn.dstFlag
        self.dataOut.errorCount = self.dataIn.errorCount
        self.dataOut.useLocalTime = self.dataIn.useLocalTime

        self.dataOut.radarControllerHeaderObj = self.dataIn.radarControllerHeaderObj.copy()
        self.dataOut.systemHeaderObj = self.dataIn.systemHeaderObj.copy()
        self.dataOut.ippSeconds = self.dataIn.getDeltaH()*(10**-6)/0.15

        self.dataOut.channelList = self.dataIn.channelList
        self.dataOut.heightList = self.dataIn.heightList
        self.dataOut.dtype = numpy.dtype([('real','<f4'),('imag','<f4')])

        self.dataOut.nBaud = self.dataIn.nBaud
        self.dataOut.nCode = self.dataIn.nCode
        self.dataOut.code = self.dataIn.code
#         self.dataOut.nProfiles = self.dataOut.nFFTPoints

        self.dataOut.flagDiscontinuousBlock = self.dataIn.flagDiscontinuousBlock
        self.dataOut.utctime = self.firstdatatime
        self.dataOut.flagDecodeData = self.dataIn.flagDecodeData #asumo q la data esta decodificada
        self.dataOut.flagDeflipData = self.dataIn.flagDeflipData #asumo q la data esta sin flip
        self.dataOut.flagShiftFFT = False

        self.dataOut.nCohInt = self.dataIn.nCohInt
        self.dataOut.nIncohInt = 1

        self.dataOut.windowOfFilter = self.dataIn.windowOfFilter

        self.dataOut.frequency = self.dataIn.frequency
        self.dataOut.realtime = self.dataIn.realtime

        self.dataOut.azimuth = self.dataIn.azimuth
        self.dataOut.zenith = self.dataIn.zenith

        self.dataOut.beam.codeList = self.dataIn.beam.codeList
        self.dataOut.beam.azimuthList = self.dataIn.beam.azimuthList
        self.dataOut.beam.zenithList = self.dataIn.beam.zenithList

    def __createLagsBlock(self, voltages):

        if self.__lags_buffer is None:
            self.__lags_buffer = numpy.zeros((self.dataOut.nChannels, self.dataOut.nProfiles, self.dataOut.nHeights), dtype='complex')

        nsegments = self.dataOut.nHeights - self.dataOut.nProfiles

#         codes = numpy.conjugate(self.__input_buffer[:,9:169])/10000

        for i in range(nsegments):
            self.__lags_buffer[:,:,i] = voltages[:,i:i+self.dataOut.nProfiles]#*codes

        return self.__lags_buffer

    def __decodeData(self, volt_buffer, pulseIndex=None):

        if pulseIndex is None:
            return volt_buffer

        codes = numpy.conjugate(self.__input_buffer[:,pulseIndex[0]:pulseIndex[1]])/10000

        nsegments = self.dataOut.nHeights - self.dataOut.nProfiles

        for i in range(nsegments):
            volt_buffer[:,:,i] = volt_buffer[:,:,i]*codes

        return volt_buffer

    def __getFft(self, datablock):
        """
        Convierte valores de Voltaje a Spectra

        Affected:
            self.dataOut.data_spc
            self.dataOut.data_cspc
            self.dataOut.data_dc
            self.dataOut.heightList
            self.profIndex
            self.__input_buffer
            self.dataOut.flagNoData
        """

        fft_volt = numpy.fft.fft(datablock, n=self.dataOut.nFFTPoints, axis=1)

        dc = fft_volt[:,0,:]

        #calculo de self-spectra
        fft_volt = numpy.fft.fftshift(fft_volt, axes=(1,))
        spc = fft_volt * numpy.conjugate(fft_volt)
        spc = spc.real

        blocksize = 0
        blocksize += dc.size
        blocksize += spc.size

        cspc = None
        pairIndex = 0

        if self.dataOut.pairsList != []:
            #calculo de cross-spectra
            cspc = numpy.zeros((self.dataOut.nPairs, self.dataOut.nFFTPoints, self.dataOut.nHeights), dtype='complex')
            for pair in self.dataOut.pairsList:
                if pair[0] not in self.dataOut.channelList:
                    raise ValueError("Error getting CrossSpectra: pair 0 of %s is not in channelList = %s" %(str(pair), str(self.dataOut.channelList)))
                if pair[1] not in self.dataOut.channelList:
                    raise ValueError("Error getting CrossSpectra: pair 1 of %s is not in channelList = %s" %(str(pair), str(self.dataOut.channelList)))

                chan_index0 = self.dataOut.channelList.index(pair[0])
                chan_index1 = self.dataOut.channelList.index(pair[1])

                cspc[pairIndex,:,:] = fft_volt[chan_index0,:,:] * numpy.conjugate(fft_volt[chan_index1,:,:])
                pairIndex += 1
            blocksize += cspc.size

        self.dataOut.data_spc = spc
        self.dataOut.data_cspc = cspc
        self.dataOut.data_dc = dc
        self.dataOut.blockSize = blocksize
        self.dataOut.flagShiftFFT = True

    def run(self, nProfiles=None, nFFTPoints=None, pairsList=[], code=None, nCode=None, nBaud=None, codeFromHeader=False, pulseIndex=None):

        self.dataOut.flagNoData = True

        self.code = None

        if codeFromHeader:
            if self.dataIn.code is not None:
                self.code = self.dataIn.code

        if code is not None:
            self.code = numpy.array(code).reshape(nCode,nBaud)

        if self.dataIn.type == "Voltage":

            if nFFTPoints == None:
                raise ValueError("This SpectraProc.run() need nFFTPoints input variable")

            if nProfiles == None:
                nProfiles = nFFTPoints

            self.profIndex == nProfiles
            self.firstdatatime = self.dataIn.utctime

            self.dataOut.ippFactor = 1
            self.dataOut.nFFTPoints = nFFTPoints
            self.dataOut.nProfiles = nProfiles
            self.dataOut.pairsList = pairsList

            self.__updateSpecFromVoltage()

            if not self.dataIn.flagDataAsBlock:
                self.__input_buffer = self.dataIn.data.copy()

                lags_block = self.__createLagsBlock(self.__input_buffer)

                lags_block = self.__decodeData(lags_block, pulseIndex)

            else:
                self.__input_buffer = self.dataIn.data.copy()

            self.__getFft(lags_block)

            self.dataOut.flagNoData = False

            return True

        raise ValueError("The type of input object '%s' is not valid"%(self.dataIn.type))

    def __selectPairs(self, pairsList):

        if channelList == None:
            return

        pairsIndexListSelected = []

        for thisPair in pairsList:

            if thisPair not in self.dataOut.pairsList:
                continue

            pairIndex = self.dataOut.pairsList.index(thisPair)

            pairsIndexListSelected.append(pairIndex)

        if not pairsIndexListSelected:
            self.dataOut.data_cspc = None
            self.dataOut.pairsList = []
            return

        self.dataOut.data_cspc = self.dataOut.data_cspc[pairsIndexListSelected]
        self.dataOut.pairsList = [self.dataOut.pairsList[i] for i in pairsIndexListSelected]

        return

    def __selectPairsByChannel(self, channelList=None):

        if channelList == None:
            return

        pairsIndexListSelected = []
        for pairIndex in self.dataOut.pairsIndexList:
            #First pair
            if self.dataOut.pairsList[pairIndex][0] not in channelList:
                continue
            #Second pair
            if self.dataOut.pairsList[pairIndex][1] not in channelList:
                continue

            pairsIndexListSelected.append(pairIndex)

        if not pairsIndexListSelected:
            self.dataOut.data_cspc = None
            self.dataOut.pairsList = []
            return

        self.dataOut.data_cspc = self.dataOut.data_cspc[pairsIndexListSelected]
        self.dataOut.pairsList = [self.dataOut.pairsList[i] for i in pairsIndexListSelected]

        return

    def selectChannels(self, channelList):

        channelIndexList = []

        for channel in channelList:
            if channel not in self.dataOut.channelList:
                raise ValueError("Error selecting channels, Channel %d is not valid.\nAvailable channels = %s" %(channel, str(self.dataOut.channelList)))

            index = self.dataOut.channelList.index(channel)
            channelIndexList.append(index)

        self.selectChannelsByIndex(channelIndexList)

    def selectChannelsByIndex(self, channelIndexList):
        """
        Selecciona un bloque de datos en base a canales segun el channelIndexList

        Input:
            channelIndexList    :    lista sencilla de canales a seleccionar por ej. [2,3,7]

        Affected:
            self.dataOut.data_spc
            self.dataOut.channelIndexList
            self.dataOut.nChannels

        Return:
            None
        """

        for channelIndex in channelIndexList:
            if channelIndex not in self.dataOut.channelIndexList:
                raise ValueError("Error selecting channels: The value %d in channelIndexList is not valid.\nAvailable channel indexes = " %(channelIndex, self.dataOut.channelIndexList))

#         nChannels = len(channelIndexList)

        data_spc = self.dataOut.data_spc[channelIndexList,:]
        data_dc = self.dataOut.data_dc[channelIndexList,:]

        self.dataOut.data_spc = data_spc
        self.dataOut.data_dc = data_dc

        self.dataOut.channelList = [self.dataOut.channelList[i] for i in channelIndexList]
#        self.dataOut.nChannels = nChannels

        self.__selectPairsByChannel(self.dataOut.channelList)

        return 1

    def selectHeights(self, minHei, maxHei):
        """
        Selecciona un bloque de datos en base a un grupo de valores de alturas segun el rango
        minHei <= height <= maxHei

        Input:
            minHei    :    valor minimo de altura a considerar
            maxHei    :    valor maximo de altura a considerar

        Affected:
            Indirectamente son cambiados varios valores a travez del metodo selectHeightsByIndex

        Return:
            1 si el metodo se ejecuto con exito caso contrario devuelve 0
        """

        if (minHei > maxHei):
            raise ValueError("Error selecting heights: Height range (%d,%d) is not valid" % (minHei, maxHei))

        if (minHei < self.dataOut.heightList[0]):
            minHei = self.dataOut.heightList[0]

        if (maxHei > self.dataOut.heightList[-1]):
            maxHei = self.dataOut.heightList[-1]

        minIndex = 0
        maxIndex = 0
        heights = self.dataOut.heightList

        inda = numpy.where(heights >= minHei)
        indb = numpy.where(heights <= maxHei)

        try:
            minIndex = inda[0][0]
        except:
            minIndex = 0

        try:
            maxIndex = indb[0][-1]
        except:
            maxIndex = len(heights)

        self.selectHeightsByIndex(minIndex, maxIndex)

        return 1

    def getBeaconSignal(self, tauindex = 0, channelindex = 0, hei_ref=None):
        newheis = numpy.where(self.dataOut.heightList>self.dataOut.radarControllerHeaderObj.Taus[tauindex])

        if hei_ref != None:
            newheis = numpy.where(self.dataOut.heightList>hei_ref)

        minIndex = min(newheis[0])
        maxIndex = max(newheis[0])
        data_spc = self.dataOut.data_spc[:,:,minIndex:maxIndex+1]
        heightList = self.dataOut.heightList[minIndex:maxIndex+1]

        # determina indices
        nheis = int(self.dataOut.radarControllerHeaderObj.txB/(self.dataOut.heightList[1]-self.dataOut.heightList[0]))
        avg_dB = 10*numpy.log10(numpy.sum(data_spc[channelindex,:,:],axis=0))
        beacon_dB = numpy.sort(avg_dB)[-nheis:]
        beacon_heiIndexList = []
        for val in avg_dB.tolist():
            if val >= beacon_dB[0]:
                beacon_heiIndexList.append(avg_dB.tolist().index(val))

        #data_spc = data_spc[:,:,beacon_heiIndexList]
        data_cspc = None
        if self.dataOut.data_cspc is not None:
            data_cspc = self.dataOut.data_cspc[:,:,minIndex:maxIndex+1]
            #data_cspc = data_cspc[:,:,beacon_heiIndexList]

        data_dc = None
        if self.dataOut.data_dc is not None:
            data_dc = self.dataOut.data_dc[:,minIndex:maxIndex+1]
            #data_dc = data_dc[:,beacon_heiIndexList]

        self.dataOut.data_spc = data_spc
        self.dataOut.data_cspc = data_cspc
        self.dataOut.data_dc = data_dc
        self.dataOut.heightList = heightList
        self.dataOut.beacon_heiIndexList = beacon_heiIndexList

        return 1


    def selectHeightsByIndex(self, minIndex, maxIndex):
        """
        Selecciona un bloque de datos en base a un grupo indices de alturas segun el rango
        minIndex <= index <= maxIndex

        Input:
            minIndex    :    valor de indice minimo de altura a considerar
            maxIndex    :    valor de indice maximo de altura a considerar

        Affected:
            self.dataOut.data_spc
            self.dataOut.data_cspc
            self.dataOut.data_dc
            self.dataOut.heightList

        Return:
            1 si el metodo se ejecuto con exito caso contrario devuelve 0
        """

        if (minIndex < 0) or (minIndex > maxIndex):
            raise ValueError("Error selecting heights: Index range (%d,%d) is not valid" % (minIndex, maxIndex))

        if (maxIndex >= self.dataOut.nHeights):
            maxIndex = self.dataOut.nHeights-1

        #Spectra
        data_spc = self.dataOut.data_spc[:,:,minIndex:maxIndex+1]

        data_cspc = None
        if self.dataOut.data_cspc is not None:
            data_cspc = self.dataOut.data_cspc[:,:,minIndex:maxIndex+1]

        data_dc = None
        if self.dataOut.data_dc is not None:
            data_dc = self.dataOut.data_dc[:,minIndex:maxIndex+1]

        self.dataOut.data_spc = data_spc
        self.dataOut.data_cspc = data_cspc
        self.dataOut.data_dc = data_dc

        self.dataOut.heightList = self.dataOut.heightList[minIndex:maxIndex+1]

        return 1

    def removeDC(self, mode = 2):
        jspectra = self.dataOut.data_spc
        jcspectra = self.dataOut.data_cspc


        num_chan = jspectra.shape[0]
        num_hei = jspectra.shape[2]

        if jcspectra is not None:
            jcspectraExist = True
            num_pairs = jcspectra.shape[0]
        else:   jcspectraExist = False

        freq_dc = jspectra.shape[1]/2
        ind_vel = numpy.array([-2,-1,1,2]) + freq_dc

        if ind_vel[0]<0:
            ind_vel[list(range(0,1))] = ind_vel[list(range(0,1))] + self.num_prof

        if mode == 1:
            jspectra[:,freq_dc,:] = (jspectra[:,ind_vel[1],:] + jspectra[:,ind_vel[2],:])/2 #CORRECCION

            if jcspectraExist:
                jcspectra[:,freq_dc,:] = (jcspectra[:,ind_vel[1],:] + jcspectra[:,ind_vel[2],:])/2

        if mode == 2:

            vel = numpy.array([-2,-1,1,2])
            xx = numpy.zeros([4,4])

            for fil in range(4):
                xx[fil,:] = vel[fil]**numpy.asarray(list(range(4)))

            xx_inv = numpy.linalg.inv(xx)
            xx_aux = xx_inv[0,:]

            for ich in range(num_chan):
                yy = jspectra[ich,ind_vel,:]
                jspectra[ich,freq_dc,:] = numpy.dot(xx_aux,yy)

                junkid = jspectra[ich,freq_dc,:]<=0
                cjunkid = sum(junkid)

                if cjunkid.any():
                    jspectra[ich,freq_dc,junkid.nonzero()] = (jspectra[ich,ind_vel[1],junkid] + jspectra[ich,ind_vel[2],junkid])/2

            if jcspectraExist:
                for ip in range(num_pairs):
                    yy = jcspectra[ip,ind_vel,:]
                    jcspectra[ip,freq_dc,:] = numpy.dot(xx_aux,yy)


        self.dataOut.data_spc = jspectra
        self.dataOut.data_cspc = jcspectra

        return 1

    def removeInterference(self,  interf = 2,hei_interf = None, nhei_interf = None, offhei_interf = None):

        jspectra = self.dataOut.data_spc
        jcspectra = self.dataOut.data_cspc
        jnoise = self.dataOut.getNoise()
        num_incoh = self.dataOut.nIncohInt

        num_channel  = jspectra.shape[0]
        num_prof  = jspectra.shape[1]
        num_hei   = jspectra.shape[2]

        #hei_interf
        if hei_interf is None:
            count_hei = num_hei/2   #Como es entero no importa
            hei_interf = numpy.asmatrix(list(range(count_hei))) + num_hei - count_hei
            hei_interf = numpy.asarray(hei_interf)[0]
        #nhei_interf
        if (nhei_interf == None):
            nhei_interf = 5
        if (nhei_interf < 1):
            nhei_interf = 1
        if (nhei_interf > count_hei):
            nhei_interf = count_hei
        if (offhei_interf == None):
            offhei_interf = 0

        ind_hei = list(range(num_hei))
#         mask_prof = numpy.asarray(range(num_prof - 2)) + 1
#         mask_prof[range(num_prof/2 - 1,len(mask_prof))] += 1
        mask_prof = numpy.asarray(list(range(num_prof)))
        num_mask_prof = mask_prof.size
        comp_mask_prof = [0, num_prof/2]


        #noise_exist:    Determina si la variable jnoise ha sido definida y contiene la informacion del ruido de cada canal
        if (jnoise.size < num_channel or numpy.isnan(jnoise).any()):
            jnoise = numpy.nan
        noise_exist = jnoise[0] < numpy.Inf

        #Subrutina de Remocion de la Interferencia
        for ich in range(num_channel):
            #Se ordena los espectros segun su potencia (menor a mayor)
            power = jspectra[ich,mask_prof,:]
            power = power[:,hei_interf]
            power = power.sum(axis = 0)
            psort = power.ravel().argsort()

            #Se estima la interferencia promedio en los Espectros de Potencia empleando
            junkspc_interf = jspectra[ich,:,hei_interf[psort[list(range(offhei_interf, nhei_interf + offhei_interf))]]]

            if noise_exist:
            #    tmp_noise = jnoise[ich] / num_prof
                tmp_noise = jnoise[ich]
            junkspc_interf = junkspc_interf - tmp_noise
            #junkspc_interf[:,comp_mask_prof] = 0

            jspc_interf = junkspc_interf.sum(axis = 0) / nhei_interf
            jspc_interf = jspc_interf.transpose()
            #Calculando el espectro de interferencia promedio
            noiseid =  numpy.where(jspc_interf <= tmp_noise/ numpy.sqrt(num_incoh))
            noiseid = noiseid[0]
            cnoiseid = noiseid.size
            interfid = numpy.where(jspc_interf > tmp_noise/ numpy.sqrt(num_incoh))
            interfid = interfid[0]
            cinterfid = interfid.size

            if (cnoiseid > 0):   jspc_interf[noiseid] = 0

            #Expandiendo los perfiles a limpiar
            if (cinterfid > 0):
                new_interfid = (numpy.r_[interfid - 1, interfid, interfid + 1] + num_prof)%num_prof
                new_interfid = numpy.asarray(new_interfid)
                new_interfid = {x for x in new_interfid}
                new_interfid = numpy.array(list(new_interfid))
                new_cinterfid = new_interfid.size
            else: new_cinterfid = 0

            for ip in range(new_cinterfid):
                ind = junkspc_interf[:,new_interfid[ip]].ravel().argsort()
                jspc_interf[new_interfid[ip]] = junkspc_interf[ind[nhei_interf/2],new_interfid[ip]]


            jspectra[ich,:,ind_hei] = jspectra[ich,:,ind_hei] - jspc_interf #Corregir indices

            #Removiendo la interferencia del punto de mayor interferencia
            ListAux = jspc_interf[mask_prof].tolist()
            maxid = ListAux.index(max(ListAux))


            if cinterfid > 0:
                for ip in range(cinterfid*(interf == 2) - 1):
                    ind = (jspectra[ich,interfid[ip],:] < tmp_noise*(1 + 1/numpy.sqrt(num_incoh))).nonzero()
                    cind = len(ind)

                    if (cind > 0):
                        jspectra[ich,interfid[ip],ind] = tmp_noise*(1 + (numpy.random.uniform(cind) - 0.5)/numpy.sqrt(num_incoh))

                ind = numpy.array([-2,-1,1,2])
                xx = numpy.zeros([4,4])

                for id1 in range(4):
                    xx[:,id1] = ind[id1]**numpy.asarray(list(range(4)))

                xx_inv = numpy.linalg.inv(xx)
                xx = xx_inv[:,0]
                ind = (ind + maxid + num_mask_prof)%num_mask_prof
                yy = jspectra[ich,mask_prof[ind],:]
                jspectra[ich,mask_prof[maxid],:] = numpy.dot(yy.transpose(),xx)


            indAux = (jspectra[ich,:,:] < tmp_noise*(1-1/numpy.sqrt(num_incoh))).nonzero()
            jspectra[ich,indAux[0],indAux[1]] = tmp_noise * (1 - 1/numpy.sqrt(num_incoh))

        #Remocion de Interferencia en el Cross Spectra
        if jcspectra is None: return jspectra, jcspectra
        num_pairs = jcspectra.size/(num_prof*num_hei)
        jcspectra = jcspectra.reshape(num_pairs, num_prof, num_hei)

        for ip in range(num_pairs):

            #-------------------------------------------

            cspower = numpy.abs(jcspectra[ip,mask_prof,:])
            cspower = cspower[:,hei_interf]
            cspower = cspower.sum(axis = 0)

            cspsort = cspower.ravel().argsort()
            junkcspc_interf = jcspectra[ip,:,hei_interf[cspsort[list(range(offhei_interf, nhei_interf + offhei_interf))]]]
            junkcspc_interf = junkcspc_interf.transpose()
            jcspc_interf = junkcspc_interf.sum(axis = 1)/nhei_interf

            ind = numpy.abs(jcspc_interf[mask_prof]).ravel().argsort()

            median_real = numpy.median(numpy.real(junkcspc_interf[mask_prof[ind[list(range(3*num_prof/4))]],:]))
            median_imag = numpy.median(numpy.imag(junkcspc_interf[mask_prof[ind[list(range(3*num_prof/4))]],:]))
            junkcspc_interf[comp_mask_prof,:] = numpy.complex(median_real, median_imag)

            for iprof in range(num_prof):
                ind = numpy.abs(junkcspc_interf[iprof,:]).ravel().argsort()
                jcspc_interf[iprof] = junkcspc_interf[iprof, ind[nhei_interf/2]]

            #Removiendo la Interferencia
            jcspectra[ip,:,ind_hei] = jcspectra[ip,:,ind_hei] - jcspc_interf

            ListAux = numpy.abs(jcspc_interf[mask_prof]).tolist()
            maxid = ListAux.index(max(ListAux))

            ind = numpy.array([-2,-1,1,2])
            xx = numpy.zeros([4,4])

            for id1 in range(4):
                xx[:,id1] = ind[id1]**numpy.asarray(list(range(4)))

            xx_inv = numpy.linalg.inv(xx)
            xx = xx_inv[:,0]

            ind = (ind + maxid + num_mask_prof)%num_mask_prof
            yy = jcspectra[ip,mask_prof[ind],:]
            jcspectra[ip,mask_prof[maxid],:] = numpy.dot(yy.transpose(),xx)

        #Guardar Resultados
        self.dataOut.data_spc = jspectra
        self.dataOut.data_cspc = jcspectra

        return 1

    def setRadarFrequency(self, frequency=None):

        if frequency != None:
            self.dataOut.frequency = frequency

        return 1

    def getNoise(self, minHei=None, maxHei=None, minVel=None, maxVel=None):
        #validacion de rango
        if minHei == None:
            minHei = self.dataOut.heightList[0]

        if maxHei == None:
            maxHei = self.dataOut.heightList[-1]

        if (minHei < self.dataOut.heightList[0]) or (minHei > maxHei):
            print('minHei: %.2f is out of the heights range'%(minHei))
            print('minHei is setting to %.2f'%(self.dataOut.heightList[0]))
            minHei = self.dataOut.heightList[0]

        if (maxHei > self.dataOut.heightList[-1]) or (maxHei < minHei):
            print('maxHei: %.2f is out of the heights range'%(maxHei))
            print('maxHei is setting to %.2f'%(self.dataOut.heightList[-1]))
            maxHei = self.dataOut.heightList[-1]

        # validacion de velocidades
        velrange = self.dataOut.getVelRange(1)

        if minVel == None:
            minVel = velrange[0]

        if maxVel == None:
            maxVel = velrange[-1]

        if (minVel < velrange[0]) or (minVel > maxVel):
            print('minVel: %.2f is out of the velocity range'%(minVel))
            print('minVel is setting to %.2f'%(velrange[0]))
            minVel = velrange[0]

        if (maxVel > velrange[-1]) or (maxVel < minVel):
            print('maxVel: %.2f is out of the velocity range'%(maxVel))
            print('maxVel is setting to %.2f'%(velrange[-1]))
            maxVel = velrange[-1]

        # seleccion de indices para rango
        minIndex = 0
        maxIndex = 0
        heights = self.dataOut.heightList

        inda = numpy.where(heights >= minHei)
        indb = numpy.where(heights <= maxHei)

        try:
            minIndex = inda[0][0]
        except:
            minIndex = 0

        try:
            maxIndex = indb[0][-1]
        except:
            maxIndex = len(heights)

        if (minIndex < 0) or (minIndex > maxIndex):
            raise ValueError("some value in (%d,%d) is not valid" % (minIndex, maxIndex))

        if (maxIndex >= self.dataOut.nHeights):
            maxIndex = self.dataOut.nHeights-1

        # seleccion de indices para velocidades
        indminvel = numpy.where(velrange >= minVel)
        indmaxvel = numpy.where(velrange <= maxVel)
        try:
            minIndexVel = indminvel[0][0]
        except:
            minIndexVel = 0

        try:
            maxIndexVel = indmaxvel[0][-1]
        except:
            maxIndexVel = len(velrange)

        #seleccion del espectro
        data_spc = self.dataOut.data_spc[:,minIndexVel:maxIndexVel+1,minIndex:maxIndex+1]
        #estimacion de ruido
        noise = numpy.zeros(self.dataOut.nChannels)

        for channel in range(self.dataOut.nChannels):
            daux = data_spc[channel,:,:]
            noise[channel] = hildebrand_sekhon(daux, self.dataOut.nIncohInt)

        self.dataOut.noise_estimation = noise.copy()

        return 1


'''
Created on Jul 3, 2014

@author: roj-idl71
'''
import os
import datetime
import numpy

try:
    from gevent import sleep
except:
    from time import sleep

from schainpy.model.data.jroheaderIO import RadarControllerHeader, SystemHeader
from schainpy.model.data.jrodata import Voltage
from schainpy.model.proc.jroproc_base import ProcessingUnit, Operation, MPDecorator

try:
    import digital_rf_hdf5
except:
    pass

class USRPReader(ProcessingUnit):
    '''
    classdocs
    '''

    def __init__(self, **kwargs):
        '''
        Constructor
        '''

        ProcessingUnit.__init__(self, **kwargs)

        self.dataOut = Voltage()
        self.__printInfo = True
        self.__flagDiscontinuousBlock = False
        self.__bufferIndex = 9999999

        self.__ippKm = None
        self.__codeType = 0
        self.__nCode = None
        self.__nBaud = None
        self.__code = None

    def __getCurrentSecond(self):

        return self.__thisUnixSample/self.__sample_rate

    thisSecond = property(__getCurrentSecond, "I'm the 'thisSecond' property.")

    def __setFileHeader(self):
        '''
        In this method will be initialized every parameter of dataOut object (header, no data)
        '''
        ippSeconds = 1.0*self.__nSamples/self.__sample_rate

        nProfiles = 1.0/ippSeconds  #Number of profiles in one second

        self.dataOut.radarControllerHeaderObj = RadarControllerHeader(ipp=self.__ippKm,
                                                                      txA=0,
                                                                      txB=0,
                                                                      nWindows=1,
                                                                      nHeights=self.__nSamples,
                                                                      firstHeight=self.__firstHeigth,
                                                                      deltaHeight=self.__deltaHeigth,
                                                                      codeType=self.__codeType,
                                                                      nCode=self.__nCode, nBaud=self.__nBaud,
                                                                      code = self.__code)

        self.dataOut.systemHeaderObj = SystemHeader(nSamples=self.__nSamples,
                                                    nProfiles=nProfiles,
                                                    nChannels=len(self.__channelList),
                                                    adcResolution=14)

        self.dataOut.type = "Voltage"

        self.dataOut.data = None

        self.dataOut.dtype = numpy.dtype([('real','<i8'),('imag','<i8')])

#        self.dataOut.nChannels = 0

#        self.dataOut.nHeights = 0

        self.dataOut.nProfiles = nProfiles

        self.dataOut.heightList = self.__firstHeigth + numpy.arange(self.__nSamples, dtype = numpy.float)*self.__deltaHeigth

        self.dataOut.channelList = self.__channelList

        self.dataOut.blocksize = self.dataOut.getNChannels() * self.dataOut.getNHeights()

#        self.dataOut.channelIndexList = None

        self.dataOut.flagNoData = True

        #Set to TRUE if the data is discontinuous
        self.dataOut.flagDiscontinuousBlock = False

        self.dataOut.utctime = None

        self.dataOut.timeZone = self.__timezone/60  #timezone like jroheader, difference in minutes between UTC and localtime

        self.dataOut.dstFlag = 0

        self.dataOut.errorCount = 0

        self.dataOut.nCohInt = 1

        self.dataOut.flagDecodeData = False #asumo que la data esta decodificada

        self.dataOut.flagDeflipData = False #asumo que la data esta sin flip

        self.dataOut.flagShiftFFT = False

        self.dataOut.ippSeconds = ippSeconds

        #Time interval between profiles
        #self.dataOut.timeInterval = self.dataOut.ippSeconds * self.dataOut.nCohInt

        self.dataOut.frequency = self.__frequency

        self.dataOut.realtime = self.__online

    def findDatafiles(self, path, startDate=None, endDate=None):

        if not os.path.isdir(path):
            return []

        try:
            digitalReadObj = digital_rf_hdf5.read_hdf5(path, load_all_metadata=True)
        except:
            digitalReadObj = digital_rf_hdf5.read_hdf5(path)

        channelNameList = digitalReadObj.get_channels()

        if not channelNameList:
            return []

        metadata_dict = digitalReadObj.get_rf_file_metadata(channelNameList[0])

        sample_rate = metadata_dict['sample_rate'][0]

        this_metadata_file = digitalReadObj.get_metadata(channelNameList[0])

        try:
            timezone = this_metadata_file['timezone'].value
        except:
            timezone = 0

        startUTCSecond, endUTCSecond = digitalReadObj.get_bounds(channelNameList[0])/sample_rate - timezone

        startDatetime = datetime.datetime.utcfromtimestamp(startUTCSecond)
        endDatatime = datetime.datetime.utcfromtimestamp(endUTCSecond)

        if not startDate:
            startDate = startDatetime.date()

        if not endDate:
            endDate = endDatatime.date()

        dateList = []

        thisDatetime = startDatetime

        while(thisDatetime<=endDatatime):

            thisDate = thisDatetime.date()

            if thisDate < startDate:
                continue

            if thisDate > endDate:
                break

            dateList.append(thisDate)
            thisDatetime += datetime.timedelta(1)

        return dateList

    def setup(self, path = None,
                    startDate = None,
                    endDate = None,
                    startTime = datetime.time(0,0,0),
                    endTime = datetime.time(23,59,59),
                    channelList = None,
                    nSamples = None,
                    ippKm = 60,
                    online = False,
                    delay = 60,
                    buffer_size = 1024,
                    **kwargs):
        '''
        In this method we should set all initial parameters.

        Inputs:
            path
            startDate
            endDate
            startTime
            endTime
            set
            expLabel
            ext
            online
            delay
        '''

        if not os.path.isdir(path):
            raise ValueError("[Reading] Directory %s does not exist" %path)

        try:
            self.digitalReadObj = digital_rf_hdf5.read_hdf5(path, load_all_metadata=True)
        except:
            self.digitalReadObj = digital_rf_hdf5.read_hdf5(path)

        channelNameList = self.digitalReadObj.get_channels()

        if not channelNameList:
            raise ValueError("[Reading] Directory %s does not have any files" %path)

        if not channelList:
            channelList = list(range(len(channelNameList)))

        ##########  Reading metadata ######################

        metadata_dict = self.digitalReadObj.get_rf_file_metadata(channelNameList[channelList[0]])

        self.__sample_rate = metadata_dict['sample_rate'][0]
#         self.__samples_per_file = metadata_dict['samples_per_file'][0]
        self.__deltaHeigth = 1e6*0.15/self.__sample_rate

        this_metadata_file = self.digitalReadObj.get_metadata(channelNameList[channelList[0]])

        self.__frequency = None
        try:
            self.__frequency = this_metadata_file['center_frequencies'].value
        except:
            self.__frequency = this_metadata_file['fc'].value

        if not self.__frequency:
            raise ValueError("Center Frequency is not defined in metadata file")

        try:
            self.__timezone = this_metadata_file['timezone'].value
        except:
            self.__timezone = 0

        self.__firstHeigth = 0

        try:
            codeType = this_metadata_file['codeType'].value
        except:
            codeType = 0

        nCode = 1
        nBaud = 1
        code = numpy.ones((nCode, nBaud), dtype=numpy.int)

        if codeType:
            nCode = this_metadata_file['nCode'].value
            nBaud = this_metadata_file['nBaud'].value
            code = this_metadata_file['code'].value

        if not ippKm:
            try:
                #seconds to km
                ippKm = 1e6*0.15*this_metadata_file['ipp'].value
            except:
                ippKm = None

        ####################################################
        startUTCSecond = None
        endUTCSecond = None

        if startDate:
            startDatetime = datetime.datetime.combine(startDate, startTime)
            startUTCSecond = (startDatetime-datetime.datetime(1970,1,1)).total_seconds() + self.__timezone

        if endDate:
            endDatetime = datetime.datetime.combine(endDate, endTime)
            endUTCSecond = (endDatetime-datetime.datetime(1970,1,1)).total_seconds() + self.__timezone

        start_index, end_index = self.digitalReadObj.get_bounds(channelNameList[channelList[0]])

        if not startUTCSecond:
            startUTCSecond = start_index/self.__sample_rate

        if start_index > startUTCSecond*self.__sample_rate:
            startUTCSecond = start_index/self.__sample_rate

        if not endUTCSecond:
            endUTCSecond = end_index/self.__sample_rate

        if end_index < endUTCSecond*self.__sample_rate:
            endUTCSecond = end_index/self.__sample_rate

        if not nSamples:
            if not ippKm:
                raise ValueError("[Reading] nSamples or ippKm should be defined")

            nSamples = int(ippKm / (1e6*0.15/self.__sample_rate))

        channelBoundList = []
        channelNameListFiltered = []

        for thisIndexChannel in channelList:
            thisChannelName =  channelNameList[thisIndexChannel]
            start_index, end_index = self.digitalReadObj.get_bounds(thisChannelName)
            channelBoundList.append((start_index, end_index))
            channelNameListFiltered.append(thisChannelName)

        self.profileIndex = 0

        self.__delay = delay
        self.__ippKm = ippKm
        self.__codeType = codeType
        self.__nCode = nCode
        self.__nBaud = nBaud
        self.__code = code

        self.__datapath = path
        self.__online = online
        self.__channelList = channelList
        self.__channelNameList = channelNameListFiltered
        self.__channelBoundList = channelBoundList
        self.__nSamples = nSamples
        self.__samples_to_read = int(buffer_size*nSamples)
        self.__nChannels = len(self.__channelList)

        self.__startUTCSecond = startUTCSecond
        self.__endUTCSecond = endUTCSecond

        self.__timeInterval = 1.0 * self.__samples_to_read/self.__sample_rate #Time interval

        if online:
#             self.__thisUnixSample = int(endUTCSecond*self.__sample_rate - 4*self.__samples_to_read)
            startUTCSecond = numpy.floor(endUTCSecond)

        self.__thisUnixSample = int(startUTCSecond*self.__sample_rate) - self.__samples_to_read

        self.__data_buffer = numpy.zeros((self.__nChannels, self.__samples_to_read), dtype = numpy.complex)

        self.__setFileHeader()
        self.isConfig = True

        print("[Reading] USRP Data was found from %s to %s " %(
                                                      datetime.datetime.utcfromtimestamp(self.__startUTCSecond - self.__timezone),
                                                      datetime.datetime.utcfromtimestamp(self.__endUTCSecond - self.__timezone)
                                                      ))

        print("[Reading] Starting process from %s to %s" %(datetime.datetime.utcfromtimestamp(startUTCSecond - self.__timezone),
                                                           datetime.datetime.utcfromtimestamp(endUTCSecond - self.__timezone)
                                                           ))

    def __reload(self):

        if not self.__online:
            return

#         print
#         print "%s not in range [%s, %s]" %(
#                                           datetime.datetime.utcfromtimestamp(self.thisSecond - self.__timezone),
#                                           datetime.datetime.utcfromtimestamp(self.__startUTCSecond - self.__timezone),
#                                           datetime.datetime.utcfromtimestamp(self.__endUTCSecond - self.__timezone)
#                                           )
        print("[Reading] reloading metadata ...")

        try:
            self.digitalReadObj.reload(complete_update=True)
        except:
            self.digitalReadObj.reload()

        start_index, end_index = self.digitalReadObj.get_bounds(self.__channelNameList[self.__channelList[0]])

        if start_index > self.__startUTCSecond*self.__sample_rate:
            self.__startUTCSecond = 1.0*start_index/self.__sample_rate

        if end_index > self.__endUTCSecond*self.__sample_rate:
            self.__endUTCSecond = 1.0*end_index/self.__sample_rate
            print()
            print("[Reading] New timerange found [%s, %s] " %(
                                                      datetime.datetime.utcfromtimestamp(self.__startUTCSecond - self.__timezone),
                                                      datetime.datetime.utcfromtimestamp(self.__endUTCSecond - self.__timezone)
                                                      ))

            return True

        return False

    def __readNextBlock(self, seconds=30, volt_scale = 218776):
        '''
        '''

        #Set the next data
        self.__flagDiscontinuousBlock = False
        self.__thisUnixSample += self.__samples_to_read

        if self.__thisUnixSample + 2*self.__samples_to_read > self.__endUTCSecond*self.__sample_rate:
            print("[Reading] There are no more data into selected time-range")

            self.__reload()

            if self.__thisUnixSample + 2*self.__samples_to_read > self.__endUTCSecond*self.__sample_rate:
                self.__thisUnixSample -=  self.__samples_to_read
                return False

        indexChannel = 0

        dataOk = False

        for thisChannelName in self.__channelNameList:

            try:
                result = self.digitalReadObj.read_vector_c81d(self.__thisUnixSample,
                                                              self.__samples_to_read,
                                                              thisChannelName)

            except IOError as e:
                #read next profile
                self.__flagDiscontinuousBlock = True
                print("[Reading] %s" %datetime.datetime.utcfromtimestamp(self.thisSecond - self.__timezone), e)
                break

            if result.shape[0] != self.__samples_to_read:
                self.__flagDiscontinuousBlock = True
                print("[Reading] %s: Too few samples were found, just %d/%d  samples" %(datetime.datetime.utcfromtimestamp(self.thisSecond - self.__timezone),
                                                                                result.shape[0],
                                                                                self.__samples_to_read))
                break

            self.__data_buffer[indexChannel,:] = result*volt_scale

            indexChannel += 1

            dataOk = True

        self.__utctime = self.__thisUnixSample/self.__sample_rate

        if not dataOk:
            return False

        print("[Reading] %s: %d samples <> %f sec" %(datetime.datetime.utcfromtimestamp(self.thisSecond - self.__timezone),
                                                         self.__samples_to_read,
                                                         self.__timeInterval))

        self.__bufferIndex = 0

        return True

    def __isBufferEmpty(self):

        if self.__bufferIndex <= self.__samples_to_read - self.__nSamples:
            return False

        return True

    def getData(self, seconds=30, nTries=5):

        '''
        This method gets the data from files and put the data into the dataOut object

        In addition, increase el the buffer counter in one.

        Return:
            data    :    retorna un perfil de voltages (alturas * canales) copiados desde el
                         buffer. Si no hay mas archivos a leer retorna None.

        Affected:
            self.dataOut
            self.profileIndex
            self.flagDiscontinuousBlock
            self.flagIsNewBlock
        '''

        err_counter = 0
        self.dataOut.flagNoData = True

        if self.__isBufferEmpty():

            self.__flagDiscontinuousBlock = False

            while True:
                if self.__readNextBlock():
                    break

                if self.__thisUnixSample > self.__endUTCSecond*self.__sample_rate:
                    return False

                if self.__flagDiscontinuousBlock:
                    print('[Reading] discontinuous block found ... continue with the next block')
                    continue

                if not self.__online:
                    return False

                err_counter += 1
                if err_counter > nTries:
                    return False

                print('[Reading] waiting %d seconds to read a new block' %seconds)
                sleep(seconds)

        self.dataOut.data = self.__data_buffer[:,self.__bufferIndex:self.__bufferIndex+self.__nSamples]
        self.dataOut.utctime = (self.__thisUnixSample + self.__bufferIndex)/self.__sample_rate
        self.dataOut.flagNoData = False
        self.dataOut.flagDiscontinuousBlock = self.__flagDiscontinuousBlock
        self.dataOut.profileIndex = self.profileIndex

        self.__bufferIndex += self.__nSamples
        self.profileIndex += 1

        if self.profileIndex == self.dataOut.nProfiles:
            self.profileIndex = 0

        return True

    def printInfo(self):
        '''
        '''
        if self.__printInfo == False:
            return

#         self.systemHeaderObj.printInfo()
#         self.radarControllerHeaderObj.printInfo()

        self.__printInfo = False

    def printNumberOfBlock(self):
        '''
        '''

        print(self.profileIndex)

    def run(self, **kwargs):
        '''
        This method will be called many times so here you should put all your code
        '''

        if not self.isConfig:
            self.setup(**kwargs)

        self.getData(seconds=self.__delay)

        return


@MPDecorator
class USRPWriter(Operation):
    '''
    classdocs
    '''

    def __init__(self, **kwargs):
        '''
        Constructor
        '''
        Operation.__init__(self, **kwargs)
        self.dataOut = None

    def setup(self, dataIn, path, blocksPerFile, set=0, ext=None):
        '''
        In this method we should set all initial parameters.

        Input:
            dataIn        :        Input data will also be outputa data

        '''
        self.dataOut = dataIn





        self.isConfig = True

        return

    def run(self, dataIn, **kwargs):
        '''
        This method will be called many times so here you should put all your code

        Inputs:

            dataIn        :        object with the data

        '''

        if not self.isConfig:
            self.setup(dataIn, **kwargs)


if __name__ == '__main__':

    readObj = USRPReader()

    while True:
        readObj.run(path='/Volumes/DATA/haystack/passive_radar/')
#         readObj.printInfo()
        readObj.printNumberOfBlock()